"""
Schematics are simplified representations of networks, intended to explain their structure and make the way they operate
understandable. The arcgis.schematics module contains the types and functions for working with schematic layers and
datasets.

"""

from arcgis.gis import Layer

"""This class provides access to diagrams and schematic layers, as well as diagram templates."""


class SchematicLayers(Layer):
    def __init__(self, url, gis=None):
        super(SchematicLayers, self).__init__(url, gis)
        try:
            from arcgis.gis.server._service._adminfactory import AdminServiceGen

            self.service = AdminServiceGen(service=self, gis=gis)
        except:
            pass

    @property
    def diagrams(self):
        """
        The Schematic Diagrams resource represents all the schematic diagrams
        under a schematic service. It is returned as an array of `Schematic
        Diagram resources <https://developers.arcgis.com/rest/services-reference/enterprise/schematic-diagram.htm>`_
        by the REST API.
        """
        params = {"f": "json"}
        exportURL = self._url + "/diagrams"
        return self._con.get(path=exportURL, params=params, token=self._token)

    # ----------------------------------------------------------------------
    @property
    def folders(self):
        """
        The Schematic Folders resource represents the set of schematic folders
        in the schematic dataset(s) related to the schematic layers under a
        schematic service. It is returned as an array of `Schematic Folder Objects <https://developers.arcgis.com/documentation/common-data-types/schematic-folder-object.htm>`_
        by the REST API.
        """
        params = {"f": "json"}
        exportURL = self._url + "/folders"
        return self._con.get(path=exportURL, params=params, token=self._token)

    # ----------------------------------------------------------------------
    @property
    def layers(self):
        """
        The Schematic Layers resource represents all the schematic layers
        under a schematic service published by ArcGIS Server. It is returned
        as an array of `Schematic Layer resources <https://developers.arcgis.com/rest/services-reference/enterprise/schematic-layer.htm>`_
        by the REST API.
        """
        params = {"f": "json"}
        exportURL = self._url + "/schematicLayers"
        return self._con.get(path=exportURL, params=params, token=self._token)

    # ----------------------------------------------------------------------
    @property
    def templates(self):
        """
        The Schematic Diagram Templates represents all the schematic diagram
        templates related to the published schematic layers under a schematic
        service. It is returned as an array of `Schematic Diagram Template
        resources <https://developers.arcgis.com/rest/services-reference/enterprise/schematic-diagram-template.htm>`_
        by the REST API.
        """
        params = {"f": "json"}
        exportURL = self._url + "/templates"
        return self._con.get(path=exportURL, params=params, token=self._token)

    # ----------------------------------------------------------------------
    def search_diagrams(
        self, whereClause=None, relatedObjects=None, relatedSchematicObjects=None
    ):
        """
        The Schematic Search Diagrams operation is performed on the schematic
        service resource. The result of this operation is an array of Schematic
        Diagram Information Object.

        It is used to search diagrams in the schematic service by criteria;
        that is, diagrams filtered out via a where clause on any schematic
        diagram class table field, diagrams that contain schematic features
        associated with a specific set of GIS features/objects, or diagrams
        that contain schematic features associated with the same GIS features/
        objects related to another set of schematic features.

        See `Schematic Search Diagrams <https://developers.arcgis.com/rest/services-reference/enterprise/schematic-search-diagrams.htm>`_
        for full details.

        =======================      =======================================================================
        **Parameter**                 **Description**
        -----------------------      -----------------------------------------------------------------------
        whereClause                  A where clause for the query filter. Any legal SQL where clause
                                     operating on the fields in the schematic diagram class table is allowed.
                                     See the `Schematic diagram class fields  <https://developers.arcgis.com/rest/services-reference/enterprise/schematic-search-diagrams.htm>`_
                                     table for the exact list of field names that can be used in this where
                                     clause.
        -----------------------      -----------------------------------------------------------------------
        relatedObjects               An array containing the list of the GIS features/objects IDs per
                                     feature class/table name that are in relation with schematic features
                                     in the resulting queried diagrams. Each GIS feature/object ID
                                     corresponds to a value of the OBJECTID field in the GIS feature
                                     class/table.
        -----------------------      -----------------------------------------------------------------------
        relatedSchematicObjects      An array containing the list of the schematic feature names per
                                     schematic feature class ID that have the same associated GIS
                                     features/objects with schematic features in the resulting queried
                                     diagrams. Each schematic feature name corresponds to a value of the
                                     ``SCHEMATICTID`` field in the schematic feature class.
        =======================      =======================================================================
        """
        params = {"f": "json"}
        if whereClause:
            params["where"] = whereClause
        if relatedObjects:
            params["relatedObjects"] = relatedObjects
        if relatedSchematicObjects:
            params["relatedSchematicObjects"] = relatedSchematicObjects

        exportURL = self._url + "/searchDiagrams"
        return self._con.get(path=exportURL, params=params, token=self._token)



from __future__ import annotations
import json
from arcgis.gis.kubernetes._admin._base import _BaseKube
from ._deployment import DeploymentManager
from ._upgrades import UpgradeManager
from ._recovery import RecoveryManager
from ._content import LanguageManager, ExternalContentManager
from ._architecture import ArchitectureManager
from ._tasks import TaskManager
from ._adaptors import WebAdaptorManager
from ._license import LicenseManager
from typing import List, Dict, Any, Tuple, Optional


class Server(_BaseKube):
    """Represents a single kubernetes service"""

    # ----------------------------------------------------------------------
    def update(self, server_json):
        """
        Updates the server JSON

        :return: Boolean

        """
        url = f"{self._url}/edit"
        params = {"f": "json", "serverJson": server_json}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    @property
    def status(self):
        """returns the status of the server"""
        url = f"{self._url}/status"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    def stop(self):
        """returns the status of the server"""
        url = f"{self._url}/stop"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def start(self):
        """returns the status of the server"""
        url = f"{self._url}/start"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res


###########################################################################
class ServerDefaults(_BaseKube):
    """Represents the server default values"""

    @property
    def properties(self):
        url = self._url
        params = {"f": "json"}
        return self._con.get(url, params)

    @properties.setter
    def properties(self, value):
        if self.properties != value:
            url = self._url + "/edit"
            params = {"f": "json", "propertyJson": value}
            self._con.post(url, params)


###########################################################################
class ServerManager(_BaseKube):
    """Manages the Registered Servers"""

    _gis = None
    _con = None
    _properties = None

    # ----------------------------------------------------------------------
    @property
    def list(self):
        """
        Allows for the modification of Servers

        :return: List of Server objects

        """
        servers = []
        if "servers" in self.properties:
            for server in self.properties.servers:
                url = f"{self._url}/{server.id}"
                servers.append(Server(url, self._gis))
        return servers

    # ----------------------------------------------------------------------
    @property
    def defaults(self):
        """
        Returns the default properties for each server type.

        :return: list of `ServerDefault`

        """
        d = []
        url = f"{self._url}/properties"
        params = {"f": "json"}
        res = self._con.get(url, params)
        if "properties" in res:
            for i in res["properties"]:
                purl = f"{url}/{i['id']}"
                d.append(ServerDefaults(url=purl, gis=self._gis))
        return d


###########################################################################
class Indexer(_BaseKube):
    """
    This resource contains connection information to the default indexing service.
    """

    def reconfigure(self) -> bool:
        """
        This operation recreates the index service metadata, schema, and data in the event it becomes corrupted.

        :returns: Boolean
        """
        params = {"f": "json"}
        url = f"{self._url}/reconfigure"
        res = self._con.post(url, params)
        return res.get("status", "failed") == "success"

    @property
    def status(self):
        """
        `status` allows you to view the status of the indexing service. You
        can view the number of users, groups, and search items in both the
        database (store) and the index. If the database and index do not
        match, indexing is either in progress or there is a problem with
        the index. It is recommended that you reindex to correct any
        issues. If indexing is in progress, you can monitor the status by
        refreshing the page.

        :return: dict

        """
        params = {"f": "json"}
        url = f"{self._url}/status"
        return self._con.get(url, params)

    def reindex(self, mode, includes=None):
        """
        The operation allows you to generate or update the indexes for content, such as users, groups, and items stored in the database store.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        mode                Required String. The mode in which the indexer should run.
                            Values: USER_MODE, GROUP_MODE, SEARCH_MODE, or FULL_MODE
        ---------------     --------------------------------------------------------------------
        includes            Optional String. A comma separated list of elements to include in
                            the index. This is useful if you want to only index certain items
                            or user accounts.
        ===============     ====================================================================

        :return: Boolean

        """
        url = f"{self._url}/reindex"
        params = {"f": "json", "mode": mode, "includes": includes}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res


class Container:
    """
    A single representation of a registered container.
    """

    _properties = None
    _url = None
    _con = None
    _gis = None

    def __init__(self, url: str, gis: "GIS"):
        self._url = url
        self._gis = gis
        self._con = gis._con

    def properties(self):
        """returns the properties of the endpoint"""
        return self._con.get(self._url, {"f": "json"})

    def edit(self, value: dict) -> dict:
        """
        Allows certain container registry properties to be updated after your organization has been configured.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required dict. A dictionary of registry properties.
        ===============     ====================================================================



        :returns: bool
        """
        url = f"{self._url}/edit"
        params = {"f": "json", "containerRegistryJson": value}
        return self._con.post(url, params).get("status", "failed") == "success"


###########################################################################
class SystemManager(_BaseKube):
    """
    The system resource is a collection of system-wide resources for a
    deployment, such as the configuration store and deployment-wide
    security.
    """

    _recovery = None
    _indexer = None
    _sm = None
    _deployments = None
    _upgrades = None
    _license = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, initialize=False):
        """Constructor


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object..
        ==================     ====================================================================

        """
        super(SystemManager, self).__init__(gis=gis, url=url)
        self._url = url
        self._gis = gis
        self._con = gis._con
        if initialize:
            self._init(gis._con)

    # ----------------------------------------------------------------------
    @property
    def containers(self) -> list:
        """
        Returns the registered containers.

        :returns: list
        """
        url = f"{self._url}/containerregistries"
        params = {"f": "json"}
        return [
            Container(url=f"{url}/{container['id']}", gis=self._gis)
            for container in self._con.get(url, params).get("containerRegistries", [])
        ]

    # ----------------------------------------------------------------------
    @property
    def deployments(self) -> DeploymentManager:
        """Manages the deployment settings for enterprise"""
        url = f"{self._url}/deployments"
        if self._deployments is None:
            self._deployments = DeploymentManager(url=url, gis=self._gis)
        return self._deployments

    # ----------------------------------------------------------------------
    @property
    def upgrades(self) -> UpgradeManager:
        """
        The upgrades resource provides access to child operations and
        resources that are used to manage the release and patch updates that
        can be applied to an ArcGIS Enterprise on Kubernetes deployment.
        During the upgrade process, this resource returns detailed,
        real-time messages regarding the status and progress of the
        upgrade. While an upgrade is in progress, child operations and
        resources for this resource will be inaccessible. Once completed,
        all child endpoints will be operational, and the JSON view of the
        resource will contain a log of the job messages and the completion
        status.
        """
        url = f"{self._url}/upgrades"
        if self._upgrades is None:
            self._upgrades = UpgradeManager(url=url, gis=self._gis)
        return self._upgrades

    # ----------------------------------------------------------------------
    @property
    def recovery(self) -> RecoveryManager:
        """
        This resource allows an administrator the ability to manage
        disaster recovery settings.

        :return: RecoveryManager
        """
        if self._recovery is None:
            url = f"{self._url}/disasterrecovery"
            self._recovery = RecoveryManager(url=url, gis=self._gis)
        return self._recovery

    # ----------------------------------------------------------------------
    @property
    def web_adaptors(self) -> WebAdaptorManager:
        """
        The webadaptors resource lists the ArcGIS Enterprise on Kubernetes Web Adaptor configured your deployment. The web adaptor can be configured using the config operation.
        """
        url = f"{self._url}/webadaptors"
        return WebAdaptorManager(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def licenses(self) -> List[Dict[str, Any]]:
        """
        The licenses resource lists the current license level of ArcGIS Server and all authorized extensions.

        :return: List[Dict[str, Any]]
        """
        url = f"{self._url}/licenses"
        return LicenseManager(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def external_content(self) -> ExternalContentManager:
        """ """
        return ExternalContentManager(url=f"{self._url}/content", gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def language(self) -> LanguageManager:
        """
        The content resource provides access to the languages resource.
        The languages resource provides a list of current languages for an
        organization.

        :return: LanguageManager

        """
        return LanguageManager(url=f"{self._url}/content", gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def tasks(self) -> TaskManager:
        """
        This resource returns a list of tasks (CleanGPJobs, BackupRetentionCleaner, CreateBackup) that exist within your deployment.
        """
        url = f"{self._url}/tasks"
        return TaskManager(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def architecture_profiles(self) -> ArchitectureManager:
        """
        This resource returns the architecture profile that is set when an organization is configured and provides access to all three architecture
        profile resources: development, standard-availability, and enhanced-availability.
        """
        url = f"{self._url}/architectureprofiles"
        return ArchitectureManager(url=url, gis=self._gis)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        Gets/Sets the system properties resource list system properties
        that have been modified to control the portal's environment.

        :return: dict
        """
        url = f"{self._url}/properties"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @properties.setter
    def properties(self, value):
        """
        Gets/Sets the system properties resource list system properties
        that have been modified to control the portal's environment.

        :return: dict
        """
        url = f"{self._url}/properties/update"
        params = {"f": "json", "properties": value}
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    @property
    def indexer(self):
        """
        Allows user to manage the site's indexer

        :return: `Indexer`
        """
        if self._indexer is None:
            url = f"{self._url}/indexer"
            self._indexer = Indexer(url=url, gis=self._gis)
        return self._indexer

    # ----------------------------------------------------------------------
    @property
    def servers(self) -> ServerManager:
        """Returns a manager to work with ArcGIS Servers registerd with Kubernetes"""
        if self._sm is None:
            self._sm = ServerManager(url=f"{self._url}/servers", gis=self._gis)
        return self._sm



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# training work but not infrencing due to transforms

model = dict(
    type="CornerNet",
    backbone=dict(
        type="HourglassNet",
        downsample_times=5,
        num_stacks=2,
        stage_channels=[256, 256, 384, 384, 384, 512],
        stage_blocks=[2, 2, 2, 2, 2, 4],
        norm_cfg=dict(type="BN", requires_grad=True),
    ),
    neck=None,
    bbox_head=dict(
        type="CentripetalHead",
        num_classes=80,
        in_channels=256,
        num_feat_levels=2,
        corner_emb_channels=0,
        loss_heatmap=dict(
            type="GaussianFocalLoss", alpha=2.0, gamma=4.0, loss_weight=1
        ),
        loss_offset=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1),
        loss_guiding_shift=dict(type="SmoothL1Loss", beta=1.0, loss_weight=0.05),
        loss_centripetal_shift=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1),
    ),
    # training and testing settings
    train_cfg=None,
    test_cfg=dict(
        corner_topk=100,
        local_maximum_kernel=3,
        distance_threshold=0.5,
        score_thr=0.05,
        max_per_img=100,
        nms_cfg=dict(type="soft_nms", iou_threshold=0.5, method="gaussian"),
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/centripetalnet/centripetalnet_hourglass104_mstest_16x6_210e_coco/centripetalnet_hourglass104_mstest_16x6_210e_coco_20200915_204804-3ccc61e5.pth"



from __future__ import annotations
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap


class BaseMissionServer(object):
    """
    Base Class for all Mission Server Classes
    """

    _url = None
    _gis = None
    _con = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis=None):
        self._url = url
        if gis is None:
            from arcgis import env

            gis = env.active_gis
        if gis is None:
            raise ValueError("A GIS could not be obtained.")
        self._gis = gis
        self._con = self._gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"< {self.__class__.__name__} @ {self._url} >"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        self._properties = None
        res = self._con.get(self._url, {"f": "json"})
        self._properties = PropertyMap(res)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """gets the service properties"""
        if self._properties is None:
            self._init()
        return self._properties



"""

.. module:: _vectortile.py
   :platform: Windows, Linux
   :synopsis: Represents functions/classes that represents a vector tile
              service.
.. moduleauthor:: Esri

"""
from __future__ import absolute_import
import tempfile
from typing import Optional
from .._common._base import BaseServer


########################################################################
class VectorTile(BaseServer):
    """
    The vector tile service resource represents a vector service published
    with ArcGIS Server. The resource provides information about the service
    such as the tile info, spatial reference, initial and full extents.
    """

    _con = None
    _url = None
    _json = None
    _json_dict = None

    def __init__(self, url, connection=None, initialize=True):
        """class initializer"""
        super(VectorTile, self).__init__(url=url, connection=connection)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)

    # ----------------------------------------------------------------------
    def tile_fonts(
        self, fontstack: str, stack_range: str, out_folder: Optional[str] = None
    ):
        """This resource returns glyphs in PBF format. The template url for
        this fonts resource is represented in Vector Tile Style resource."""
        url = "{url}/resources/fonts/{fontstack}/{stack_range}.pbf".format(
            url=self._url, fontstack=fontstack, stack_range=stack_range
        )
        params = {}
        if out_folder is None:
            out_folder = tempfile.gettempdir()
        return self._con.get(path=url, params=params, out_folder=out_folder)

    # ----------------------------------------------------------------------
    def vector_tile(
        self, level: str, row: str, column: str, out_folder: Optional[str] = None
    ):
        """This resource represents a single vector tile for the map. The
        bytes for the tile at the specified level, row and column are
        returned in PBF format. If a tile is not found, an HTTP status code
        of 404 (Not found) is returned."""
        url = "{url}/tile/{level}/{row}/{column}.pbf".format(
            url=self._url, level=level, row=row, column=column
        )
        params = {}

        if out_folder is None:
            out_folder = tempfile.gettempdir()
        return self._con.get(
            path=url, params=params, out_folder=out_folder, try_json=False
        )

    # ----------------------------------------------------------------------
    def tile_sprite(
        self, out_format: str = "sprite.json", out_folder: Optional[str] = None
    ):
        """
        This resource returns sprite image and metadata
        """
        url = "{url}/resources/sprites/{f}".format(url=self._url, f=out_format)
        if out_folder is None:
            out_folder = tempfile.gettempdir()
        return self._con.get(path=url, params={}, out_folder=out_folder)

    # ----------------------------------------------------------------------
    @property
    def info(self):
        """This returns relative paths to a list of resource files"""
        url = "{url}/resources/info".format(url=self._url)
        params = {"f": "json"}
        return self._con.get(path=url, params=params)

    # ----------------------------------------------------------------------
    @property
    def iteminfo(self):
        """returns the item information for the service"""
        url = "{url}/resources/info/iteminfo".format(url=self._url)
        params = {"f": "json"}
        return self._con.get(path=url, params=params)

    # ----------------------------------------------------------------------
    @property
    def metadata(self):
        """returns the item's metadata for the service"""
        url = "{url}/resources/info/metedata".format(url=self._url)
        params = {"f": "json"}
        return self._con.get(path=url, params=params)

    # ----------------------------------------------------------------------
    @property
    def thumbnail(self):
        """returns the item's thumbnail for the service"""
        url = "{url}/resources/info/thumbnail".format(url=self._url)
        params = {"f": "json"}
        return self._con.get(path=url, params=params)



from ._search import _search
from ._certificates import CertificateManager
from ._jb import StatusJob
from ._apikeys import APIKeyManager, APIKey
from ._dataclasses import (
    ItemTypeEnum,
    ItemProperties,
    CreateServiceParameter,
    MetadataFormatEnum,
    ServiceTypeEnum,
    SpatialFilter,
    SpatialRelationship,
    ViewLayerDefParameter,
)



from __future__ import annotations
import os
import uuid
import json
import shutil
import zipfile
import tempfile
import warnings
import concurrent.futures
from typing import Any

from arcgis._impl.common._utils import _date_handler
from arcgis.auth.tools import LazyLoader

_arcgis = LazyLoader("arcgis")
_arcgis_ux = LazyLoader("arcgis.gis.admin._ux")
__all__ = ["UXCloner"]


class UXCloner:
    """
    The `UXCloner` class facilitates the copying of UX settings from one WebGIS site to another,
    enabling administrators to maintain a consistent user experience across multiple sites.
    The cloning operations can be performed either online, through site-to-site transfer via HTTP,
    or offline, using a file package.

    ================    ===============================================================
    **Parameter**       **Description**
    ----------------    ---------------------------------------------------------------
    gis                 Required GIS. The source WebGIS.
    ================    ===============================================================

    """

    _source_gis: _arcgis.GIS | None = None
    _RESOURCE_IGNORE_LIST: list[str]
    _tp: concurrent.futures.ThreadPoolExecutor

    #  --------------------------------------------------------------------
    def __init__(self, gis: _arcgis.GIS | None):
        self._source_gis = gis
        self._RESOURCE_IGNORE_LIST = [
            "localizedOrgProperties",
            "travelmodes.json",
            "Survey123Properties",
        ]
        self._tp = concurrent.futures.ThreadPoolExecutor(max_workers=10)

    #  --------------------------------------------------------------------
    def clone(
        self,
        *,
        targets: list[_arcgis.gis.GIS] | None = None,
        workspace_folder: str | None = None,
        package_save_folder: str | None = None,
        package_name: str | None = None,
    ) -> list[concurrent.futures.Future]:
        """
        Copies the UX settings from the source WebGIS site to the destination WebGIS site or to
        a `.uxpk` offline file.
        When directly connected to two WebGIS', this method performs the clone operation immediately.
        When cloning in an offlien situation, a `.uxpk` is created and stored on the user's local
        hard drive.



        :returns: list[concurrent.futures.Future]

        """
        if targets is None:
            targets = [None]
        jobs: list[concurrent.futures.Future] = []
        for target_gis in targets:
            params = {
                "target_gis": target_gis,
                "out_folder": workspace_folder,
                "package_folder": package_save_folder,
                "package_name": package_name,
            }
            jobs.append(self._tp.submit(self._clone, **params))
            # jobs.append(self._clone(**params))
            del target_gis, params
        return jobs

    #  --------------------------------------------------------------------
    def _clone(
        self,
        *,
        target_gis: _arcgis.gis.GIS | None = None,
        out_folder: str | None = None,
        package_folder: str | None = None,
        package_name: str | None = None,
    ) -> concurrent.futures.Future:
        """clones the settings from site A to site B"""
        if target_gis is None:
            if package_folder is None:
                package_folder = os.path.join(
                    tempfile.gettempdir(), uuid.uuid4().hex[:6]
                )
                os.makedirs(package_folder, exist_ok=True)
            return self._create_offline_clone_package(
                out_folder=out_folder,
                package_name=package_name,
                package_save_location=package_folder,
            )
        else:
            return self._online_clone_workflow(
                target_gis=target_gis, out_folder=out_folder
            )

    #  --------------------------------------------------------------------
    def _setup_folder_workspaces(self, out_folder: str | None = None) -> tuple[str]:
        """
        Creates the output location of the data


        :returns: tuple[str] of the workspace folder.
        The root, HomePageSettings folder and ResourceManager Folder
        """
        if out_folder is None:
            out_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex[:5])
            hp_folder = os.path.join(out_folder, "hp")
            resource_folder = os.path.join(out_folder, "resources")
            os.makedirs(out_folder, exist_ok=True)

        else:
            hp_folder = os.path.join(out_folder, "hp")
            resource_folder = os.path.join(out_folder, "resources")
            os.makedirs(out_folder, exist_ok=True)
        os.makedirs(hp_folder, exist_ok=True)
        os.makedirs(resource_folder, exist_ok=True)
        return out_folder, hp_folder, resource_folder

    #  --------------------------------------------------------------------
    def _create_configuration(
        self,
        offline: bool,
        out_folder: str,
        hp_folder: str,
        resource_folder: str,
    ) -> dict[str, Any]:
        """
        This creates the configuration object that holds all the UX configuration data.

        :returns: dict[str,Any]
        """
        source_ux: _arcgis_ux.UX = self._source_gis.admin.ux
        hp: _arcgis_ux.HomePageSettings = source_ux.homepage_settings
        mp: _arcgis_ux.MapSettings = source_ux.map_settings
        ip: _arcgis_ux.ItemSettings = source_ux.item_settings
        ss: _arcgis_ux.SecuritySettings = source_ux.security_settings
        pp: _arcgis.gis.admin.PasswordPolicy = self._source_gis.admin.password_policy
        cs: _arcgis.gis.admin.CategoryManager = self._source_gis.admin.category_schema

        settings_configuration: dict[str, Any] = {
            "offline": offline,
            "name": source_ux.name,
            "summary": source_ux.summary,
            "contact_link": source_ux.contact_link,
            "admin_contacts": source_ux.admin_contacts,
            "help_source": source_ux.help_source,
            "logo": source_ux.get_logo(out_folder),
            "description": source_ux.description,
            "description_visibility": source_ux.description_visibility,
            "HomePageSettings": {
                "resource": hp._portal_resources.get("home.page.json"),
                "resource_file_name": "home.page.json",
                "typography": hp.get_typography(),
                "background": hp.get_background(hp_folder),
                "logo": hp.get_logo(hp_folder),
            },
            "MapSettings": {
                "default_mapviewer": mp.default_mapviewer,
                "default_extent": mp.default_extent,
                "default_basemap": mp.default_basemap,
                "use_vector_basemap": mp.use_vector_basemap,
                "vector_basemap": mp.vector_basemap,
                "basemap_gallery_group": self._source_gis._get_properties(True).get(
                    "basemapGalleryGroupQuery", None
                ),
                "units": mp.units,
                "config_apps_group": self._source_gis._get_properties(True).get(
                    "templatesGroupQuery", None
                ),  # templatesGroupQuery
                "analysis_layer_group": self._source_gis._get_properties(True).get(
                    "analysisLayersGroupQuery", None
                ),
            },
            "ItemSettings": {
                "comments": ip.enable_comments,
                "enable_metadata_edit": ip.enable_metadata_edit,
                "metadata_format": ip.metadata_format,
            },
            "SecuritySettings": {
                "allowed_origins": ",".join(ss.allowed_origins),  #
                "allowed_redirect_uris": ss.allowed_redirect_uris,  #
                "enable_https": ss.enable_https,  #
                "anonymous_access": ss.anonymous_access,  #
                "enable_update_user_profile": ss.enable_update_user_profile,  #
                "share_public": ss.share_public,  #
                "can_sign_in_social": dict(self._source_gis.properties).get(
                    "canSignInSocial", False
                ),
                "show_social_media": ss.show_social_media,  #
                "signin_settings": ss.signin_settings,  #
                "trusted_servers": ss.trusted_servers,  #
                "informational_banner": ss.get_informational_banner(),  #
                "org_access_notice": ss.get_org_access_notice(),
                "anonymous_access_notice": ss.get_anonymous_access_notice(),
                "multifactor_authentication": ss.get_multifactor_authentication(),
                "email_settings": ss.get_email_settings(),
            },
            "PasswordPolicy": {
                "password_policy": pp.policy,
                "lockout_policy": pp.lockout_policy,
            },
            "CategorySchema": {"schema": cs.schema},
            "ResourceManager": {
                "resources": [],
                "folder": os.path.basename(resource_folder),
                "offline": offline,
            },
        }
        return settings_configuration

    #  --------------------------------------------------------------------
    def _download_resources(
        self, resource_folder: str, settings_configuration: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Downloads and stores the GIS resources items to a local workspace
        The config file is then updated with the files.
        :returns: dict[str,Any]
        """
        rm = self._source_gis.admin.resources
        for resource in rm.list(num=-1):
            if not resource["key"] in self._RESOURCE_IGNORE_LIST:
                del resource["created"]
                del resource["size"]
                resource["path"] = rm.get(resource["key"], resource_folder)
                settings_configuration["ResourceManager"]["resources"].append(resource)
        return settings_configuration

    #  --------------------------------------------------------------------
    def load_offline_configuration(self, package: str) -> concurrent.futures.Future:
        """
        Loads the UX configuration file into the current active portal.
        """
        workspace = os.path.join(
            tempfile.gettempdir(), f"ux_config_{uuid.uuid4().hex[:5]}"
        )
        os.makedirs(workspace, exist_ok=True)
        shutil.unpack_archive(filename=package, extract_dir=workspace, format="zip")
        hp_folder = os.path.join(workspace, "hp")
        resource_folder = os.path.join(workspace, "resources")
        out_folder = workspace
        os.makedirs(hp_folder, exist_ok=True)
        os.makedirs(resource_folder, exist_ok=True)
        os.makedirs(out_folder, exist_ok=True)
        settings_configuration = {}
        with open(
            os.path.join(workspace, "settings_configuration.json"), "r"
        ) as reader:
            settings_configuration = json.load(reader)
        if settings_configuration:
            ## 1). Update the settings_configuration for the resources
            ##
            rs = []
            for r in settings_configuration["ResourceManager"]["resources"]:
                r = dict(r)
                r["path"] = os.path.join(resource_folder, os.path.basename(r["path"]))
                rs.append(r)
            settings_configuration["ResourceManager"]["resources"] = rs
            if "logo" in settings_configuration and settings_configuration["logo"]:
                logo = os.path.join(
                    workspace,
                    os.path.basename(settings_configuration["logo"]),
                )
                settings_configuration["logo"] = logo

        return self._config_or_org(
            settings_configuration=settings_configuration,
            target_gis=self._source_gis,
            hp_folder=hp_folder,
            resource_folder=resource_folder,
            out_folder=out_folder,
        )

    #  --------------------------------------------------------------------
    def _zipdir(self, src, dst, zip_name):
        """
        Function creates zip archive from src in dst location. The name of archive is zip_name.
        :param src: Path to directory to be archived.
        :param dst: Path where archived dir will be stored.
        :param zip_name: The name of the archive.
        :return: None
        """

        zip_name = os.path.join(dst, zip_name)
        ### zipfile handler
        ziph = zipfile.ZipFile(zip_name, "w")
        ### writing content of src directory to the archive
        for root, dirs, files in os.walk(src):
            for file in files:
                ziph.write(
                    os.path.join(root, file),
                    arcname=os.path.join(root.replace(src, ""), file),
                )
        ziph.close()
        return zip_name

    #  --------------------------------------------------------------------
    def _create_cloner_file(
        self, save_folder: str, path: str, package_name: str | None = None
    ) -> str:
        """Creates the Zipped compressed .uxpk file."""
        if package_name is None:
            package_name = f"{uuid.uuid4().hex[:5]}.uxpk"
        else:
            package_name += ".uxpk"
        save_fp: str = os.path.join(save_folder, package_name)
        self._zipdir(src=path, dst=save_folder, zip_name=package_name)

        if os.path.isfile(save_fp):
            return save_fp
        else:
            raise Exception(f"Could not save the the file: {save_fp}")
        return save_fp

    #  --------------------------------------------------------------------
    def _create_offline_clone_package(
        self,
        out_folder: str | None = None,  #  workspace folder
        package_save_location: str | None = None,  # package save location
        package_name: str | None = None,  # package save name
    ) -> str:
        """
        Creates an offline cloner package (.uxpk)
        """
        offline: bool = True
        ## 1). Setup the workspace folder
        ##
        (
            out_folder,
            hp_folder,
            resource_folder,
        ) = self._setup_folder_workspaces(out_folder=out_folder)
        ## 2). Creat the configuration file.
        ##
        settings_configuration = self._create_configuration(
            offline=offline,
            out_folder=out_folder,
            hp_folder=hp_folder,
            resource_folder=resource_folder,
        )
        ## 3). Create the cloner file
        ##
        settings_configuration = self._download_resources(
            resource_folder=resource_folder,
            settings_configuration=settings_configuration,
        )
        ## 4). Save the settings_configuration to a .json file
        ##
        settings_config_file: str = os.path.join(
            out_folder, "settings_configuration.json"
        )
        with open(settings_config_file, "w") as writer:
            json.dump(settings_configuration, writer, default=_date_handler)
        ## 4). Package up the data and return it
        ##
        os.makedirs(name=package_save_location, exist_ok=True)
        save_fp: str = self._create_cloner_file(
            save_folder=package_save_location,
            path=out_folder,
            package_name=package_name,
        )
        ## 5). Clean up files and folders
        ##

        shutil.rmtree(path=hp_folder, ignore_errors=True)
        shutil.rmtree(path=resource_folder, ignore_errors=True)
        if package_save_location.lower().find(out_folder.lower()) == -1:
            shutil.rmtree(out_folder, ignore_errors=True)
        return save_fp

    def _config_or_org(
        self,
        settings_configuration: dict[str, Any],
        target_gis: _arcgis.gis.GIS,
        hp_folder: str,
        resource_folder: str,
        out_folder: str,
    ):
        """
        Moves the configuration file to the GIS
        """
        if target_gis._portal.is_arcgisonline:
            from arcgis.gis.admin import AGOLAdminManager

            admin: AGOLAdminManager = target_gis.admin

        else:
            from arcgis.gis.admin import PortalAdminManager

            admin: PortalAdminManager = target_gis.admin
        ux: _arcgis_ux.UX = admin.ux
        for resource in settings_configuration["ResourceManager"]["resources"]:
            admin.resources.add(**resource)
        if settings_configuration["logo"]:
            ux.set_logo(logo_file=settings_configuration["logo"], show_logo=True)

        ux.name = settings_configuration["name"]
        ux.summary = settings_configuration["summary"]
        if (
            "contact_link" in settings_configuration
            and settings_configuration["contact_link"]
        ):
            ux.contact_link = settings_configuration["contact_link"]["contactUs"].get(
                "url", ""
            )
        try:
            ux.admin_contacts = settings_configuration["admin_contacts"]
        except Exception as ex:
            warnings.warn(
                f"Could not find the administration user: {settings_configuration['admin_contacts']}"
            )
        try:
            ux.help_source = settings_configuration["help_source"]
        except Exception as ex:
            warnings.warn(
                f"Warning: setting the help source property could not be set: {ex}"
            )
        ux.description = settings_configuration["description"]
        ux.description_visibility = settings_configuration["description_visibility"]
        ux.item_settings.enable_comments = settings_configuration["ItemSettings"][
            "comments"
        ]
        ux.item_settings.enable_metadata_edit = settings_configuration["ItemSettings"][
            "enable_metadata_edit"
        ]
        ux.item_settings.metadata_format = settings_configuration["ItemSettings"][
            "metadata_format"
        ]
        #  Map Settings
        ux.map_settings.default_extent = settings_configuration["MapSettings"][
            "default_extent"
        ]
        if settings_configuration["MapSettings"]["default_mapviewer"]:
            ux.map_settings.default_mapviewer = settings_configuration["MapSettings"][
                "default_mapviewer"
            ]
        if settings_configuration["MapSettings"]["units"]:
            ux.map_settings.units = settings_configuration["MapSettings"]["units"]
        # Security Settings
        security_config: dict[str, Any] = settings_configuration["SecuritySettings"]
        ux.security_settings.allowed_origins = security_config["allowed_origins"]
        ux.security_settings.allowed_redirect_uris = security_config[
            "allowed_redirect_uris"
        ]
        ux.security_settings.anonymous_access = security_config["anonymous_access"]
        ux.security_settings.enable_https = security_config["enable_https"]
        ux.security_settings.enable_update_user_profile = security_config[
            "enable_update_user_profile"
        ]
        ux.security_settings.share_public = security_config["share_public"]
        ux.security_settings.show_social_media = security_config["show_social_media"]
        signin_settings = security_config["signin_settings"]
        if signin_settings:
            ux.security_settings.set_approved_apps(
                block_unapproved=signin_settings["blockUnapprovedThirdpartyApps"]
            )
            ux.security_settings.set_blocked_apps(
                block_beta_apps=signin_settings["blockBetaApps"],
                apps=signin_settings["blockedApps"],
            )
            ux.security_settings.set_social_media_login(
                social_login=security_config["can_sign_in_social"],
                social_networks=signin_settings["signinOptionsOrder"]["social"],
                social_network_order=signin_settings["signinOptionsOrder"]["logins"],
            )
        ux.security_settings.trusted_servers = security_config["trusted_servers"]
        # handle method updates
        #
        notice = security_config["anonymous_access_notice"]
        if notice and len(notice) > 0:
            ux.security_settings.set_anonymous_access_notice(
                title=notice["title"],
                text=notice["text"],
                button_type=notice["buttons"],
            )
        if security_config["informational_banner"]:
            ux.security_settings.set_informational_banner(
                text=security_config["informational_banner"]["text"],
                bg_color=security_config["informational_banner"]["bgColor"],
                font_color=security_config["informational_banner"]["fontColor"],
                enabled=security_config["informational_banner"]["enabled"],
            )
        if security_config["org_access_notice"]:
            ux.security_settings.set_org_access_notice(
                title=security_config["org_access_notice"]["title"],
                text=security_config["org_access_notice"]["text"],
                button_type=security_config["org_access_notice"]["buttons"],
            )
        # "PasswordPolicy"
        #
        if (
            "PasswordPolicy" in settings_configuration
            and settings_configuration["PasswordPolicy"]
        ):
            password_policy = settings_configuration["PasswordPolicy"]
            policy = password_policy["password_policy"]
            policy.pop("created", None)
            policy.pop("modified", None)
            policy.pop("type", None)

            target_gis.admin.password_policy.policy = policy
            lockout = password_policy["lockout_policy"]
            lockout.pop("created", None)
            lockout.pop("modified", None)
            lockout.pop("type", None)
            target_gis.admin.password_policy.lockout_policy = lockout
            del lockout, policy
        # Category Manager
        #
        category_schema = settings_configuration["CategorySchema"]
        if category_schema and category_schema["schema"]["categorySchema"]:
            target_gis.admin.category_schema.schema = category_schema["schema"][
                "categorySchema"
            ]
        ## 5). Clean up files and folders
        ##
        if hp_folder:
            shutil.rmtree(path=hp_folder, ignore_errors=True)
        if resource_folder:
            shutil.rmtree(path=resource_folder, ignore_errors=True)
        if out_folder:
            shutil.rmtree(out_folder, ignore_errors=True)
        return True

    #  --------------------------------------------------------------------
    def _online_clone_workflow(
        self,
        *,
        target_gis: _arcgis.gis.GIS,
        out_folder: str | None = None,
    ) -> "CloneJob":
        """ """
        offline: bool = False
        from arcgis.gis import User

        target_user: User = target_gis.users.me
        assert (
            target_user.role == "org_admin"
        ), "User must be an administrator to perform this operation."
        ## 1). Setup the workspace folder
        ##
        (
            out_folder,
            hp_folder,
            resource_folder,
        ) = self._setup_folder_workspaces(out_folder=out_folder)
        ## 2). Creat the configuration file.
        ##
        settings_configuration = self._create_configuration(
            offline=offline,
            out_folder=out_folder,
            hp_folder=hp_folder,
            resource_folder=resource_folder,
        )
        settings_configuration = self._download_resources(
            resource_folder=resource_folder,
            settings_configuration=settings_configuration,
        )
        ## 3). Perform the clone operations
        ##

        return self._config_or_org(
            settings_configuration=settings_configuration,
            target_gis=target_gis,
            hp_folder=hp_folder,
            resource_folder=resource_folder,
            out_folder=out_folder,
        )



""" Defines store functions for working with Attachments.
"""

from ... import workforce


def get_attachments(assignment):
    """Gets all Attachments for the assignment.
    :param assignment: An assignment that exists on the server.
    :returns: list of Attachments.
    """
    feature_layer = assignment.project.assignments_layer
    attachment_infos = feature_layer.attachments.get_list(assignment.object_id)
    return [workforce.Attachment(assignment, attinfo) for attinfo in attachment_infos]


def add_attachment(assignment, file_path):
    """Adds an Attachment to the assignment.
    :param assignment: the assignment to add the attachment to
    :param file_path: local path of file to upload.
    :raises ServerError: Indicates that the server rejected the attachment upload.
    """
    feature_layer = assignment.project.assignments_layer
    response = feature_layer.attachments.add(assignment.object_id, file_path)
    if not response["addAttachmentResult"]["success"]:
        raise workforce.ServerError([response["addAttachmentResult"]["error"]])


def delete_attachments(assignment, attachments):
    """Removes the attachments from the assignment.
    :param assignment: An assignment that exists on the server.
    :param attachments: The attachments to remove.
    :raises ServerError: Indicates that the server rejected the attachment removal.
    """
    if attachments:
        feature_layer = assignment.project.assignments_layer
        attachment_ids = ",".join([str(attachment.id) for attachment in attachments])
        response = feature_layer.attachments.delete(
            assignment.object_id, attachment_ids
        )
        errors = [
            result.error
            for result in response["deleteAttachmentResults"]
            if not result["success"]
        ]
        if errors:
            raise workforce.ServerError(errors)



"""
Mapping Holds the Plot function for creating a FeatureCollection JSON plus the render options
"""
import json
from typing import Optional, Union
import pandas as pd
import arcgis
from arcgis.mapping.renderer import (
    generate_classbreaks,
    generate_heatmap,
    generate_simple,
    generate_unique,
)
from arcgis.mapping.symbol import create_symbol, display_colormaps, show_styles
from arcgis.mapping.renderer import (
    generate_renderer,
    generate_classbreaks,
    generate_heatmap,
    generate_simple,
    generate_unique,
)
from arcgis.widgets import MapView

CLASSIFICATIONS = {
    "simple": {"renderer_type": "u"},  # simple
    "quantiles": {"renderer_type": "u"},  # quantiles
    "Quantiles": {"renderer_type": "u"},
}

CLASS_CMAPS = {
    "classless": "Greys",
    "unique_values": "Paired",
    "quantiles": "hot_r",
    "fisher_jenks": "hot_r",
    "equal_interval": "hot_r",
    "hot_spot": "hot_r",
}


RENDERER_TYPES = {
    "s": "simple",  ##
    "u": "unique",  ##
    "h": "heatmap",  ##
    "c": "ClassBreaks",  ##
    # "p" : 'Predominance',
    "str": "Stretch",  #
    "t": "Temporal",  #
    "v": "vector field",  #
}


def plot(
    df,
    map_widget: Optional[Union[arcgis.mapping.WebMap, MapView]] = None,
    name: Optional[str] = None,
    renderer_type: Optional[str] = None,
    symbol_type: Optional[str] = None,
    symbol_style: Optional[str] = None,
    col: Optional[Union[str, list]] = None,
    colors: Optional[Union[str, list, object]] = "jet",
    alpha: float = 1,
    **kwargs,
):
    """

    Plot draws the data on a web map. The user can describe in simple terms how to
    renderer spatial data using symbol.  To make the process simplier a palette
    for which colors are drawn from can be used instead of explicit colors.


    ======================  =========================================================
    **Explicit Argument**   **Description**
    ----------------------  ---------------------------------------------------------
    df                      Required Spatially Enabled DataFrame or GeoSeries. This is the data
                            to map.
    ----------------------  ---------------------------------------------------------
    map_widget              Optional WebMap object. This is the map to display the
                            data on.
    ----------------------  ---------------------------------------------------------
    name                    Optional string. The name to assign as a title of the map widget.
    ----------------------  ---------------------------------------------------------
    renderer_type           Optional string.  Determines the type of renderer to use
                            for the provided dataset. The default is 's' which is for
                            simple renderers.

                            Allowed values:

                            + 's' - is a simple renderer that uses one symbol only.
                            + 'u' - unique renderer symbolizes features based on one
                                    or more matching string attributes.
                            + 'c' - A class breaks renderer symbolizes based on the
                                    value of some numeric attribute.
                            + 'h' - heatmap renders point data into a raster
                                    visualization that emphasizes areas of higher
                                    density or weighted values.
    ----------------------  ---------------------------------------------------------
    symbol_type             Optional string. This is the type of symbol the user
                            needs to create.  Valid inputs are: simple, picture, text,
                            or carto.  The default is simple.
    ----------------------  ---------------------------------------------------------
    symbol_style            Optional string. This is the symbology used by the
                            geometry.  For example 's' for a Line geometry is a solid
                            line. And '-' is a dash line.

                            Allowed symbol types based on geometries:

                            **Point Symbols**

                             + 'o' - Circle (default)
                             + '+' - Cross
                             + 'D' - Diamond
                             + 's' - Square
                             + 'x' - X

                             **Polyline Symbols**

                             + 's' - Solid (default)
                             + '-' - Dash
                             + '-.' - Dash Dot
                             + '-..' - Dash Dot Dot
                             + '.' - Dot
                             + '--' - Long Dash
                             + '--.' - Long Dash Dot
                             + 'n' - Null
                             + 's-' - Short Dash
                             + 's-.' - Short Dash Dot
                             + 's-..' - Short Dash Dot Dot
                             + 's.' - Short Dot

                             **Polygon Symbols**

                             + 's' - Solid Fill (default)
                             + '\' - Backward Diagonal
                             + '/' - Forward Diagonal
                             + '|' - Vertical Bar
                             + '-' - Horizontal Bar
                             + 'x' - Diagonal Cross
                             + '+' - Cross

    ----------------------  ---------------------------------------------------------
    col                     Optional string/list. Field or fields used for heatmap,
                            class breaks, or unique renderers.
    ----------------------  ---------------------------------------------------------
    colors                  Optional string. The colormap, RGB array, or list of
                            either that determines the symbol color(s) for the data.
                            The default cmap is 'jet'. To get a visual representation
                            of the allowed color maps, use the **display_colormaps**
                            method.
    ----------------------  ---------------------------------------------------------
    alpha                   Optional float.  This is a value between 0 and 1 with 1
                            being the default value.  The alpha sets the transparancy
                            of the renderer when applicable.
    ======================  =========================================================

    The kwargs parameter accepts all parameters of the create_symbol method and the
    create_renderer method.


    """
    renderer = kwargs.pop("renderer", None)

    if not hasattr(df, "spatial") and not hasattr(df, "geom"):
        raise ValueError("DataFrame or Series must be spatially enabled.")

    if renderer_type is None and renderer is None and df.spatial.renderer:
        renderer = json.loads(df.spatial.renderer.json)

    if isinstance(df, pd.Series) and df.dtype.name == "geometry":
        fid = df.index.tolist()
        sdf = pd.DataFrame(data=fid, columns=["OID"])
        sdf["SHAPE"] = df
        return plot(
            df=sdf,
            map_widget=map_widget,
            name=name,
            renderer_type=renderer_type,
            symbol_type=symbol_type,
            symbol_style=symbol_style,
            col=col,
            colors=colors,
            alpha=alpha,
            **kwargs,
        )
    r = None
    if isinstance(col, str):
        col = [col]
    map_exists = True
    if symbol_type is None:
        symbol_type = "simple"
    if name is None:
        import uuid

        name = uuid.uuid4().hex[:7]
    if map_widget is None:
        map_exists = False
        map_widget = MapView()
    import string

    trantab = str.maketrans(string.punctuation, "_" * len(string.punctuation))
    col_new = [col.translate(trantab) for col in df.columns]
    col_old = df.columns.tolist()
    df.columns = col_new
    fc = df.spatial.to_feature_collection(name=name)
    df.columns = col_old
    gt = [el for el in df.spatial.geometry_type if el is not None]
    if renderer_type is None and not renderer is None:
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = renderer
        if map_exists:
            map_widget.add_layer(fc, options={"title": name})
        else:
            map_widget.add_layer(fc, options={"title": name})
            return map_widget
        return
    elif renderer_type in [None, "s"]:
        renderer_type = "s"  # simple (default)
        r = generate_simple(
            geometry_type=gt[0].lower(),
            sdf_or_series=df,
            label=name,
            symbol_type=symbol_type,
            symbol_style=symbol_style,
            colors=colors,
            alpha=alpha,
            **kwargs,
        )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r
    elif isinstance(col, str) and col not in df.columns:
        raise ValueError("Columns %s does not exist." % col)
    elif (
        isinstance(col, (tuple, list, str))
        and all([c in df.columns for c in col]) == True
        and renderer_type in ["u", "c"]
    ):
        if isinstance(col, str):
            col = [col]
        idx = 1
        if renderer_type == "u":
            for c in col:
                kwargs["field%s" % idx] = c
                idx += 1
            r = generate_unique(
                geometry_type=kwargs.pop("geometry_type", gt[0].lower()),
                sdf_or_series=df,
                symbol_type=symbol_type,
                symbol_style=symbol_style,
                colors=colors,
                alpha=alpha,
                **kwargs,
            )
        elif renderer_type == "c":
            kwargs["field"] = col[0]
            r = generate_classbreaks(
                geometry_type=kwargs.pop("geometry_type", gt[0].lower()),
                sdf_or_series=df,
                symbol_type=symbol_type,
                symbol_style=symbol_style,
                colors=colors,
                alpha=alpha,
                **kwargs,
            )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r
    elif renderer_type in ["u", "u-a"]:
        r = generate_unique(
            geometry_type=kwargs.pop("geometry_type", gt[0].lower()),
            sdf_or_series=df,
            symbol_type=symbol_type,
            symbol_style=symbol_style,
            colors=colors,
            alpha=alpha,
            **kwargs,
        )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r
    elif renderer_type == "h":
        r = generate_heatmap(
            sdf_or_series=df,
            colors=colors,
            alpha=alpha,
            blur_radius=kwargs.pop("blur_radius", 10),
            field=kwargs.pop("field", None),
            max_intensity=kwargs.pop("max_intensity", 10),
            min_intensity=kwargs.pop("min_intensity", 0),
            ratio=kwargs.pop("ratio", 0.01),
            stops=kwargs.pop("stops", 3),
            show_none=kwargs.pop("show_none", False),
        )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r
    elif renderer_type == "str":
        r = generate_renderer(
            geometry_type=kwargs.pop("geometry_type", gt[0].lower()),
            sdf_or_series=df,
            label=name,
            symbol_type=None,
            symbol_style=None,
            render_type=renderer_type,
            colors=colors,
            alpha=alpha,
            **kwargs,
        )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r
    elif renderer_type == "t":
        r = generate_renderer(
            geometry_type=kwargs.pop("geometry_type", gt[0].lower()),
            sdf_or_series=df,
            label=name,
            symbol_type=None,
            symbol_style=None,
            render_type=renderer_type,
            colors=colors,
            alpha=alpha,
            **kwargs,
        )
        fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = r

    if map_exists:
        map_widget.add_layer(fc, options={"title": name})
    else:
        map_widget.add_layer(fc, options={"title": name})
        return map_widget



from .certificate import pfx_to_pem
from .verifycontext import no_ssl_verification
from ._lazy import LazyLoader
from ._util import parse_url, assemble_url

__all__ = [
    "LazyLoader",
    "pfx_to_pem",
    "no_ssl_verification",
    "parse_url",
    "assumble_url",
]



"""
All the spatial analysis tools from the analyze_patterns, enrich_data, find_locations, manage_data, summarize_data and
use_proximity submodules, in one place for convenience.
"""

from .analyze_patterns import (
    calculate_density,
    find_hot_spots,
    find_outliers,
    find_point_clusters,
    interpolate_points,
    summarize_center_and_dispersion,
)
from .use_proximity import (
    connect_origins_to_destinations,
    create_buffers,
    create_drive_time_areas,
    find_nearest,
    plan_routes,
)
from .enrich_data import enrich_layer
from .find_locations import (
    choose_best_facilities,
    create_viewshed,
    create_watersheds,
    derive_new_locations,
    find_centroids,
    find_existing_locations,
    find_similar_locations,
    trace_downstream,
)
from .manage_data import (
    create_route_layers,
    dissolve_boundaries,
    extract_data,
    generate_tessellation,
    merge_layers,
    overlay_layers,
)
from .summarize_data import (
    aggregate_points,
    join_features,
    summarize_center_and_dispersion,
    summarize_nearby,
    summarize_within,
)



import os

from arcgis.widgets._mapview._raster._numpy_utils import *

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

from arcgis.widgets._mapview._raster.engine import RasterData


class MatplotlibRasterEngine:
    def _is_numpy_array(self, image):
        """numpy arrays behave unpredictably in `isinstance()` func calls:
        do a hacky string comparison on the type() of image arg
        """
        return "numpy" in str(type(image)) and "array" in str(type(image))

    def _get_uuid_from_numpy_array(self, image):
        return str(hash(str(image.tostring())))

    _SUPPORTED_IMAGE_FORMATS = [".png", ".jpg", ".tif"]

    def _is_supported_image_file(self, image):
        if not isinstance(image, str):
            return False
        _, ext = os.path.splitext(image)
        return os.path.exists(image) and ext in self._SUPPORTED_IMAGE_FORMATS

    def _is_arcgis_raster(self, image):
        try:
            from arcgis.raster import Raster

            return isinstance(image, Raster)
        except ImportError:
            return False

    def try_get_raster_data(self, image, bands) -> RasterData:
        if is_numpy_array(image):
            return RasterData(
                img_data=image, extent=None, id_=get_hash_numpy_array(image)
            )
        elif self._is_arcgis_raster(image):
            img_data = image.read().squeeze()
            return RasterData(
                img_data=img_data,
                extent=image.extent,
                id_=get_hash_numpy_array(img_data),
            )
        elif self._is_supported_image_file(image):
            return RasterData(
                img_data=plt.imread(image), extent=None, id_=str(hash(image))
            )
        elif isinstance(image, str) and not os.path.exists(image):
            raise Exception(f"File {image} does not exist")
        else:
            raise Exception(f"Argument {image} not supported")



from pickle import TRUE
import warnings
import traceback
from pathlib import Path
from ._codetemplate import code
from ._arcgis_model import ArcGISModel, _get_device

HAS_FASTAI = TRUE

try:
    import json
    import os
    import math
    import torch
    import PIL
    from PIL import Image
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList
    from .._video_utils import VideoUtils
    from .._utils.env import raise_fastai_import_error
    from .._utils.common import (
        get_multispectral_data_params_from_emd,
        _get_emd_path,
        read_image,
    )
    from .._image_utils import _draw_predictions
    from .._utils.env import is_arcgispronotebook
    from .._utils.object_tracking_data import get_image_for_tracking
    import matplotlib
    import matplotlib.pyplot as plt

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


try:
    import cv2

    HAS_OPENCV = TRUE
except Exception:
    HAS_OPENCV = False


class EfficientDet(ArcGISModel):
    """
    Creates a EfficientDet model for Object Detection. Supports RGB -JPEG imagery. Based on TFLite Model Maker

    =====================   =====================================================
    **Argument**            **Description**
    ---------------------   -----------------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
                            Only (JPEG+PASCAL_VOC_rectangles) format supported.
    ---------------------   -----------------------------------------------------
    backbone                Optional String. Backbone convolutional neural network
                            model used for EfficientDet.
    ---------------------   -----------------------------------------------------
    pretrained_path         Optional String. Path where a compatible pre-trained
                            model is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   =====================================================

    :return: :class:`~arcgis.learn.EfficientDet` Object
    """

    def __init__(
        self,
        data,
        backbone=None,
        pretrained_path=None,
        *args,
        **kwargs,
    ):
        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception)
        self._check_tf()

        from ._efficientdet_utils import (
            EfficientDetTrainer,
            check_data_sanity,
            _get_tf_data_loader,
        )

        self._learn_version = kwargs.get("ArcGISLearnVersion", "1.9.1")

        if self._learn_version >= "1.9.0":
            if pretrained_path is not None:
                pretrained_backbone = False
            else:
                pretrained_backbone = True
            kwargs["pretrained_backbone"] = pretrained_backbone
        else:
            del kwargs["ArcGISLearnVersion"]

        if not check_data_sanity(data, self.supported_datasets):
            raise Exception("\nInvalid data format\n")

        if not hasattr(data, "_is_empty"):
            data._is_empty = False

        super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)

        self._check_dataset_support(data)
        self._backend = "tensorflow"

        class BackBone:
            def __init__(self, default_name, backbone_name=None):
                if backbone_name is None or not type(backbone_name) is str:
                    self.__name__ = default_name
                else:
                    self.__name__ = backbone_name
                self._keras_api_names = [self.__name__]

            def get_name(self):
                return self.__name__

        self._backbone = BackBone(self._get_default_backbone(), backbone)

        if not self._check_backbone_support(self._backbone):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        self._tf_data_loader = _get_tf_data_loader(data)
        self._trainer = EfficientDetTrainer.create(data, self._backbone.get_name())
        self._tf_dataset = None
        if not self._data.is_empty:
            self._tf_dataset = self._trainer.get_tf_dataset(
                self._tf_data_loader, self._data.batch_size
            )
        else:
            self._tf_dataset = "empty"

        self._setup_backend(data, pretrained_path)

        self._code = code

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return EfficientDet._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["PASCAL_VOC_rectangles"]

    @staticmethod
    def _supported_backbones():
        return [
            "efficientdet_lite0",
            "efficientdet_lite1",
            "efficientdet_lite2",
            "efficientdet_lite3",
            "efficientdet_lite4",
        ]

    @property
    def supported_backbones(self):
        """Supported torchvision backbones for this model."""
        return EfficientDet._supported_backbones()

    def _get_default_backbone(self):
        return EfficientDet._supported_backbones()[0]

    def _setup_backend(self, data, pretrained_path):
        from ._efficientdet_utils import EfficientDetLearner
        import tensorflow as tf
        from tensorflow.keras.models import Model
        from fastai.basics import defaults

        if data._is_multispectral:
            raise Exception("Multispectral data is not supported with efficientdet")

        self._device = torch.device("cpu")
        self._data = data
        self._trainer.setup_model()
        self._opt_func = self._trainer.get_opt_func()
        self._model = self._trainer.model

        self.learn = EfficientDetLearner(
            data,
            self._model,
            opt_func=self._opt_func,
            loss_func=None,
            true_wd=True,
            bn_wd=True,
            wd=defaults.wd,
            train_bn=True,
            tf_dataset=self._tf_dataset,
            _trainer=self._trainer,
        )

        if pretrained_path is not None:
            self.load(pretrained_path)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.EfficientDet` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.EfficientDet` Object
        """
        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        backbone = emd["ModelParameters"]["backbone"]
        dataset_type = emd.get("DatasetType", "PASCAL_VOC_rectangles")
        chip_size = emd["ImageWidth"]
        resize_to = emd.get("resize_to", None)
        kwargs = emd.get("Kwargs", {})
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        data_passed = True
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                from fastai.vision import ImageList, imagenet_stats

                sd = ImageList([], path=emd_path.parent.parent.parent).split_by_idx([])
                data = (
                    sd.label_const(
                        0,
                        label_cls=ObjectDetectionCategoryList,
                        classes=list(class_mapping.values()),
                    )
                    .transform(ds_tfms)
                    .databunch(device=_get_device())
                    .normalize(imagenet_stats)
                )

            # Add 1 for background class
            data.c += 1
            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type
            data.batch_size = 1
            data.path = emd_path.parent
            data.orig_path = None
            data._val_split_pct = 0.0

        data.resize_to = resize_to
        efficientdet = cls(data, backbone, pretrained_path=str(model_file), **kwargs)

        if not data_passed:
            efficientdet.learn.data.single_ds.classes = efficientdet._data.classes
            efficientdet.learn.data.single_ds.y.classes = efficientdet._data.classes

        return efficientdet

    def _save_tflite(self, name, post_processed=True, quantized=False):
        return self.learn._save_tflite(name, post_processed, quantized)

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelType"] = "ObjectDetection"
        _emd_template["ModelName"] = type(self).__name__
        bckbn_name = self._backbone.__name__
        _emd_template["backbone"] = bckbn_name
        _emd_template["Classes"] = []

        class_data = {}
        class_data["Value"] = 0
        class_data["Name"] = "background"
        class_data["Color"] = [random.choice(range(256)) for i in range(3)]
        _emd_template["Classes"].append(class_data.copy())

        for k, class_name in self._data.class_mapping.items():
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())
        return _emd_template

    @property
    def _model_metrics(self):
        ap_dict = dict()
        if self._data._is_empty:
            import numpy as np

            ap_dict["AP"] = np.float64(0.0)
        else:
            metrics = self.learn.compute_metrics()
            mean = self.learn._compute_mean_avp
            if mean is False:
                for value in metrics:
                    if value.startswith("AP_/"):
                        class_name = value[4:]
                        ap_dict[class_name] = metrics[value]
            else:
                ap_dict = metrics["AP"]
        return dict({"average_precision_score": ap_dict})

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=True,
        visualize=False,
        resize=False,
        **kwargs,
    ):
        """
        Predicts and displays the results of a trained model on a single image.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean.
                                Will return the probability scores of the
                                bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        =====================   ===========================================

        :return: 'List' of xmin, ymin, width, height, labels, scores, of predicted bounding boxes on the given image
        """

        if not HAS_OPENCV:
            raise Exception(
                "This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24"
            )

        if isinstance(image_path, str):
            image = Image.open(image_path)
            orig_frame = cv2.imread(image_path)
        else:
            image = image_path
            orig_frame = image

        predictions, labels, scores = self.learn.predict(image, threshold, nms_overlap)

        if visualize:
            image = _draw_predictions(orig_frame, predictions, labels)
            import matplotlib.pyplot as plt

            plt.xticks([])
            plt.yticks([])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        =====================   ===========================================

        """

        VideoUtils.predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold,
            nms_overlap,
            track,
            visualize,
            output_file_path,
            multiplex,
            multiplex_file_path,
            tracker_options,
            visual_options,
            resize,
        )

    def average_precision_score(self, mean=False):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        self.learn._compute_mean_avp = mean
        avp = self._model_metrics["average_precision_score"]
        self.learn._compute_mean_avp = False
        return avp

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        """
        self._check_requisites()
        # Get Number of items
        nrows = rows
        ncols = 2
        imsize = 5
        type_data_loader = "validation"
        return_fig = False

        if not self._data._is_empty:
            x_batch = self.learn.get_gt_batches(nrows, type_data_loader)
        else:
            x_batch = None

        nrows = min(nrows, len(x_batch))

        title_font_size = 16
        top = 1 - (math.sqrt(title_font_size) / math.sqrt(100 * nrows * imsize))

        fig, ax = plt.subplots(
            nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
        )

        fig.suptitle("Ground Truth / Predictions", fontsize=title_font_size)
        if not self._data._is_empty:
            for i in range(nrows):
                if nrows == 1:
                    ax_i = ax
                else:
                    ax_i = ax[i]

                for j in range(2):
                    image = get_image_for_tracking(x_batch[i][0])
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                    if j == 0:
                        bboxes = x_batch[i][1]
                    else:
                        bboxes, _, _ = self.predict(x_batch[i][0], thresh, nms_overlap)
                    for bbox in bboxes:
                        cv2.rectangle(
                            image,
                            (int(bbox[0]), int(bbox[1])),
                            (int(bbox[2]) + int(bbox[0]), int(bbox[3] + int(bbox[1]))),
                            (255, 0, 0),
                            2,
                        )

                    ax_i[j].imshow(image)  # image
                    ax_i[j].axis("off")

        plt.subplots_adjust(top=top)
        if self._device == torch.device("cuda"):
            torch.cuda.empty_cache()
        if is_arcgispronotebook():
            plt.show()
        if return_fig:
            return fig

    def fit(
        self,
        epochs=10,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,  # "all", "best", True, False ("best" and True are same.)
        tensorboard=False,
        monitor="valid_loss",  # whatever is passed here, earlystopping and checkpointing will use that.
        **kwargs,
    ):
        """
        Train the model for the specified number of epochs and using the
        specified learning rates

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``,
                                an optimal learning rate is automatically deduced
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no
                                learning rate schedule is used.
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if parameter
                                `monitor` value stops improving for 5 epochs.
                                A minimum difference of 0.001 is required for
                                it to be considered an improvement.
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean or string.
                                Parameter to save checkpoint during training.
                                If set to `True` the best model
                                based on `monitor` will be saved during
                                training. If set to 'all', all checkpoints
                                are saved. If set to False, checkpointing will
                                be off. Setting this parameter loads the best
                                model at the end of training.Recommended to set to False.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                `<dataset-path>/training_log` which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.

                                .. note::
                                    Not applicable for Text Models
        ---------------------   -------------------------------------------
        monitor                 Optional string. Parameter specifies
                                which metric to monitor while checkpointing
                                and early stopping. Defaults to 'valid_loss'. Value
                                should be one of the metric that is displayed in
                                the training table. Use `{model_name}.available_metrics`
                                to list the available metrics to set here.
        =====================   ===========================================
        """
        if lr is None:
            lr = self._trainer.model_spec.config.learning_rate
        super().fit(
            epochs,
            lr,
            one_cycle,
            early_stopping,
            checkpoint,
            tensorboard,
            monitor,
            **kwargs,
        )



"""
These functions calculate total counts, lengths, areas, and basic descriptive statistics of features and their attributes
within areas or near other features.

aggregate_points calculates statistics about points that fall within specified areas.
summarize_nearby calculates statistics for features and their attributes that are within a specified distance.
summarize_within calculates statistics for area features and attributes that overlap each other.
"""
from __future__ import annotations
from datetime import datetime
from re import U
from typing import Any, Optional, Union

import arcgis as _arcgis
from arcgis.features.feature import FeatureCollection
from arcgis.features.layer import FeatureLayer, FeatureLayerCollection
from arcgis.gis import GIS, Item
from .._impl.common._utils import _date_handler
from .._impl.common._utils import inspect_function_inputs
import arcgis.network as network


# --------------------------------------------------------------------------
def aggregate_points(
    point_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    polygon_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ] = None,
    keep_boundaries_with_no_points: bool = True,
    summary_fields: list[str] = [],
    group_by_field: Optional[str] = None,
    minority_majority: bool = False,
    percent_points: bool = False,
    output_name: Optional[Union[str, FeatureLayer]] = None,
    context: Optional[dict] = None,
    gis: Optional[GIS] = None,
    estimate: bool = False,
    future: bool = False,
    bin_type: Optional[str] = None,
    bin_size: Optional[float] = None,
    bin_size_unit: Optional[str] = None,
):
    """
    .. image:: _static/images/agg_points_standard/aggregate_points_standard.png

    The Aggregate Points task works with a layer of point features and a layer of polygon features. It first figures out which points fall within each polygon's area.
    After determining this point-in-polygon spatial relationship, statistics about all points in the polygon are calculated and assigned to the area. The most basic statistic is the count of the number of points within the polygon, but you can get other statistics as well.

    For example, if your points represented coffee shops and each point has a TOTAL_SALES attribute, you can get statistics like the sum of all TOTAL_SALES within the polygon, or the minimum or maximum TOTAL_SALES value, or the standard deviation of all sales within the polygon.

    ====================================    ====================================================================
    **Parameter**                            **Description**
    ------------------------------------    --------------------------------------------------------------------
    point_layer                             Required point layer. The point features that will be aggregated
                                            into the polygons in the polygon_layer. See :ref:`Feature Input<FeatureInput>`.
    ------------------------------------    --------------------------------------------------------------------
    polygon_layer                           Optional polygon layer. The polygon features (areas) into which the input points will be aggregated. See :ref:`Feature Input<FeatureInput>`. The `polygon_layer` is **required** if the `bin_type`, `bin_size` and `bin_size_unit` are not specified.
    ------------------------------------    --------------------------------------------------------------------
    keep_boundaries_with_no_points          Optional boolean. A Boolean value that specifies whether the polygons that have no points within them should be returned in the output. The default is true.
    ------------------------------------    --------------------------------------------------------------------
    summary_fields                          Optional list of strings. A list of field names and statistical summary type that you wish to calculate for all points within each polygon.
                                            Note that the count of points within each polygon is always returned.
                                            summary type is one of the following:

                                            * Sum - Adds the total value of all the points in each polygon
                                            * Mean - Calculates the average of all the points in each polygon.
                                            * Min - Finds the smallest value of all the points in each polygon.
                                            * Max - Finds the largest value of all the points in each polygon.
                                            * Stddev - Finds the standard deviation of all the points in each polygon.

                                            Example [fieldName1 summaryType1,fieldName2 summaryType2].
    ------------------------------------    --------------------------------------------------------------------
    group_by_field                          Optional string. A field name in the point_layer. Points that have
                                            the same value for the group by field will have their own counts and
                                            summary field statistics. You can create statistical groups using an
                                            attribute in the analysis layer.
                                            For example, if you are aggregating crimes to neighborhood boundaries,
                                            you may have an attribute Crime_type with five different crime types.
                                            Each unique crime type forms a group, and the statistics you choose will
                                            be calculated for each unique value of Crime_type. When you choose
                                            a grouping attribute, two results are created: the result layer and a
                                            related table containing the statistics.
    ------------------------------------    --------------------------------------------------------------------
    minority_majority                       Optional boolean. This boolean parameter is applicable only when a
                                            group_by_field is specified. If true, the minority (least dominant) or
                                            the majority (most dominant) attribute values for each group field
                                            within each boundary are calculated. Two new fields are added to the
                                            aggregated_layer prefixed with `Majority_` and `Minority_`.
                                            The default is false.
    ------------------------------------    --------------------------------------------------------------------
    percent_points                          Optional boolean. This boolean parameter is applicable only when a
                                            group_by_field is specified. If set to true, the percentage count of
                                            points for each unique group_by_field value is calculated.
                                            A new field is added to the group summary output table containing the
                                            percentages of each attribute value within each group.

                                            If minority_majority is true, two additional fields are added to the
                                            aggregated_layer containing the percentages of the minority and majority
                                            attribute values within each group.
    ------------------------------------    --------------------------------------------------------------------
    output_name                             Optional string or :class:`~arcgis.features.FeatureLayer`. Existing
                                            feature layer will cause the new layer to be appended to the Feature Service.
                                            If overwrite is True in context, new layer will overwrite existing layer.
                                            If output_name not indicated then new :class:`~arcgis.features.FeatureCollection` created.
    ------------------------------------    --------------------------------------------------------------------
    context                                 Optional dict. Additional settings such as processing extent and output spatial reference.
                                            For aggregate_points, there are three settings (`overwrite` is required).

                                            - ``extent`` - a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                                            - ``outSR`` - the output features will be projected into the output spatial reference referred to by the `wkid`.
                                            - ``overwrite`` - if True, then the feature layer in output_name will be overwritten with new feature layer. Available for ArcGIS Online or Enterprise 10.9.1+

                                                .. code-block:: python

                                                    # Example Usage
                                                    context = {"extent": {"xmin": 3164569.408035,
                                                                        "ymin": -9187921.892449,
                                                                        "xmax": 3174104.927313,
                                                                        "ymax": -9175500.875353,
                                                                        "spatialReference":{"wkid":102100,"latestWkid":3857}},
                                                                "outSR": {"wkid": 3857},
                                                                "overwrite": True}
    ------------------------------------    --------------------------------------------------------------------
    gis                                     Optional, the :class:`~arcgis.gis.GIS` on which this tool runs.
                                            If not specified, the active GIS is used.
    ------------------------------------    --------------------------------------------------------------------
    estimate                                Optional Boolean. If True, the number of credits to run the operation
                                            will be returned.
    ------------------------------------    --------------------------------------------------------------------
    future                                  Optional boolean. If True, a future object will be returned and the process
                                            will not wait for the task to complete. The default is False, which means wait for results.
    ------------------------------------    --------------------------------------------------------------------
    bin_type                                Optional String. The type of bin that will be generated and points
                                            will be aggregated into. Bin options are as follows: Hexagon and Square.
                                            Square is the Default. When generating bins, for Square, the number and
                                            units specified determine the height and length of the square.
                                            For Hexagon, the number and units specified determine the distance
                                            between parallel sides. Either `bin_type` or `polygon_layer` must be
                                            specified. If `bin_type` is chosen, then `bin_size` and `bin_size_unit`
                                            specifying the size of the bins must be included.
    ------------------------------------    --------------------------------------------------------------------
    bin_size                                Optional Float. The distance for the bins of type
                                            `bin_type` that the `point_layer` will be aggregated into.
                                            When generating bins for `Square` the number and units specified determine
                                            the height and length of the square. For `Hexagon`, the number and units
                                            specified determine the distance between parallel sides.
    ------------------------------------    --------------------------------------------------------------------
    bin_size_unit                           Optional String. The linear unit to be used with the distance value
                                            specified in `bin_size`.
                                            Values: `Meters, Kilometers, Feet, Miles, NauticalMiles, or Yards`
    ====================================    ====================================================================

    :return: result_layer : :class:`~arcgis.features.FeatureLayer` if output_name is specified, else :class:`~arcgis.features.FeatureCollection`.
    If ``future = True``, then the result is a :class:`~concurrent.futures.Future` object. Call ``result()`` to get the response.


    .. code-block:: python

        USAGE EXAMPLE: To find number of permits issued in each zip code of US.

        agg_result = aggregate_points(point_layer=permits,
                                polygon_layer=zip_codes,
                                keep_boundaries_with_no_points=False,
                                summary_fields=["DeclValNu mean","DeclValNu2 mean"],
                                group_by_field='Declared_V',
                                minority_majority=True,
                                percent_points=True,
                                output_name="aggregated_permits",
                                context={"extent":{"xmin":-8609738.077325115,"ymin":4743483.445485223,"xmax":-8594030.268012533,"ymax":4752206.821338257,"spatialReference":{"wkid":102100,"latestWkid":3857}}})

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    kwargs = {
        "point_layer": point_layer,
        "polygon_layer": polygon_layer,
        "keep_boundaries_with_no_points": keep_boundaries_with_no_points,
        "summary_fields": summary_fields,
        "group_by_field": group_by_field,
        "minority_majority": minority_majority,
        "percent_points": percent_points,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "estimate": estimate,
        "future": future,
        "bin_type": bin_type,
        "bin_size": bin_size,
        "bin_size_unit": bin_size_unit,
    }
    params = inspect_function_inputs(
        fn=gis._tools.featureanalysis._tbx.aggregate_points, **kwargs
    )
    return gis._tools.featureanalysis.aggregate_points(**params)


# --------------------------------------------------------------------------
def summarize_nearby(
    sum_nearby_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    summary_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    near_type: str = "StraightLine",
    distances: Optional[list[str]] = [],
    units: str = "Meters",
    time_of_day: Optional[datetime] = None,
    time_zone_for_time_of_day: str = "GeoLocal",
    return_boundaries: bool = True,
    sum_shape: bool = True,
    shape_units: Optional[str] = None,
    summary_fields: Optional[list[str]] = [],
    group_by_field: Optional[str] = None,
    minority_majority: bool = False,
    percent_shape: bool = False,
    output_name: Optional[Union[str, FeatureLayer]] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    estimate: bool = False,
    future: bool = False,
):
    """
    .. image:: _static/images/summarize_nearby/summarize_nearby.png

    The ``summarize_nearby`` method finds features that are within a specified distance of features in the input layer.
    Distance can be measured as a straight-line distance, a drive-time distance (for example, within 10 minutes), or a
    drive distance (within 5 kilometers). Statistics are then calculated for the nearby features. For example:

    * Calculate the total population within five minutes of driving time of a proposed new store location.
    * Calculate the number of freeway access ramps within a one-mile driving distance of a proposed new store location to use as a measure of
      store accessibility.

    =========================   ====================================================================================================================
    **Parameter**                **Description**
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    sum_nearby_layer            Required :class:`~arcgis.features.FeatureLayer` . Point, line, or polygon features from which distances will be measured to features in the ``summary_layer``. See :ref:`Feature Input<FeatureInput>`.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    summary_layer               Required layer. Point, line, or polygon features. Features in this layer that are within the specified distance to features in the ``sum_nearby_layer`` will be summarized. See :ref:`Feature Input<FeatureInput>`.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    near_type                   Optional string.
                                Defines what kind of distance measurement you want to use, either straight-line distance, travel
                                time or travel distance along a street network using various modes of transportation known as travel modes.
                                The default is ``StraightLine``.

                                Choice list:

                                * ``StraightLine``,
                                * ``Driving Distance``,
                                * ``Driving Time``,
                                * ``Rural Driving Distance``,
                                * ``Rural Driving Time``,
                                * ``Trucking Distance``,
                                * ``Trucking Time``,
                                * ``Walking Distance``,
                                * ``Walking Time``
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    distances                   Optional list of float values. Defines the search distance for 'StraightLine' and distance-based travel modes, or time
                                duration for time-based travel modes. You can enter single or multiple values, separating each value with a space.
                                Features that are within (or equal to) the distances you enter will be summarized. The unit for `distances` is
                                supplied by the units parameter.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    units                       Optional string. If :attr:`near_type` is `StraightLine` or a distance-based travel mode, this is the linear unit to be
                                used with the distance value(s) specified in distances.

                                Choice list:
                                | [``Meters``, ``Kilometers``, ``Feet``, ``Yards``, ``Miles``]

                                If ``near_type`` is a time-based travel mode, the following values can be used as units:

                                Choice list:

                                | [``Seconds``, ``Minutes``, ``Hours``]

                                The default is 'Meters'.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    time_of_day                 Optional datetime.datetime. Specify whether travel times should consider traffic conditions. To use traffic in the analysis,
                                set ``near_type`` to a travel mode object whose impedance_attribute_name property is set to travel_time and assign a value
                                to ``time_of_day``. (A travel mode with other impedance_attribute_name values don't support traffic.) The ``time_of_day`` value represents
                                the time at which travel begins, or departs, from the origin points. The time is specified as datetime.datetime.

                                The service supports two kinds of traffic: typical and live. Typical traffic references travel speeds that are made up of historical
                                averages for each five-minute interval spanning a week. Live traffic retrieves speeds from a traffic feed that processes phone probe
                                records, sensors, and other data sources to record actual travel speeds and predict speeds for the near future.

                                The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries
                                Esri currently provides traffic data for.

                                Typical Traffic:

                                To ensure the task uses typical traffic in locations where it is available, choose a time and day of the week, and then convert the day
                                of the week to one of the following dates from 1990:

                                * Monday - 1/1/1990
                                * Tuesday - 1/2/1990
                                * Wednesday - 1/3/1990
                                * Thursday - 1/4/1990
                                * Friday - 1/5/1990
                                * Saturday - 1/6/1990
                                * Sunday - 1/7/1990
                                Set the time and date as datetime.datetime.

                                For example, to solve for 1:03 p.m. on Thursdays, set the time and date to 1:03 p.m., 4 January 1990; and convert to
                                datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                Live Traffic:

                                To use live traffic when and where it is available, choose a time and date and convert to datetime.

                                Esri saves live traffic data for 4 hours and references predictive data extending 4 hours into the future. If the time and date you
                                specify for this parameter is outside the 24-hour time window, or the travel time in the analysis continues past the predictive data window,
                                the task falls back to typical traffic speeds.

                                Examples:
                                from datetime import datetime

                                * "time_of_day"- datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                * "time_of_day"- datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                * "time_of_day"- datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014,
                                live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                * "time_of_day"- datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015,
                                live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    time_zone_for_time_of_day   Optional string. Specify the time zone or zones of the ``time_of_day`` parameter.

                                Choice list: ['GeoLocal', 'UTC']

                                GeoLocal-refers to the time zone in which the originsLayer points are located.

                                UTC-refers to Coordinated Universal Time.

                                The default is 'GeoLocal'.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    return_boundaries           Optional boolean. If true, the ``result_layer`` will contain areas defined by the specified ``near_type``. For example, if using 'StraightLine' of 5 miles,
                                the ``result_layer`` will contain areas with a 5 mile radius around the input ``sum_nearby_layer`` features.

                                If False, the ``result_ayer`` will contain the same features as the ``sum_nearby_layer``.

                                The default is True.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    sum_shape                   Optional boolean. A boolean value that instructs the task to calculate statistics based on shape type of the ``summary_layer``,
                                such as the length of lines or areas of polygons of the ``summary_layer`` within each polygon in ``sum_within_layer``.

                                The default is True.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    shape_units                 Optional string. If ``sum_shape`` is true, you must specify the units of the shape summary.
                                Values:

                                * When ``summary_layer`` contains polygons: Values: ['Acres', 'Hectares', 'SquareMeters', 'SquareKilometers', 'SquareFeet', 'SquareYards', 'SquareMiles']
                                * When ``summary_layer`` contains lines: Values: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    summary_fields              Optional list of strings.A list of field names and statistical summary types that you want to calculate.
                                Note that the count is always returned by default.

                                fieldName is the name of one of the numeric fields found in the input join layer.

                                statisticType is one of the following:

                                * ``SUM``-Adds the total value of all the points in each polygon
                                * ``MEAN``-Calculates the average of all the points in each polygon
                                * ``MIN``-Finds the smallest value of all the points in each polygon
                                * ``MAX``-Finds the largest value of all the points in each polygon
                                * ``STDDEV``-Finds the standard deviation of all the points in each polygon

                                Example: ["fieldName summaryType","fieldName summaryType", ...]
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    group_by_field              Optional string. This is a field of the ``summary_layer`` features that you can use to calculate statistics separately for each unique attribute value.
                                For example, suppose the ``summary_layer`` contains point locations of businesses that store hazardous materials, and one of the fields is HazardClass
                                containing codes that describe the type of hazardous material stored. To calculate summaries by each unique value of HazardClass, use HazardClass as
                                the ``group_by_field`` field.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    minority_majority           Optional boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified. If true, the minority (least dominant) or the
                                majority (most dominant) attribute values for each group field within each nearby area are calculated. Two new fields are added to
                                the ``result_layer`` prefixed with `Majority_` and `Minority_`.

                                The default is False.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    percent_shape               Optional boolean. This Boolean parameter is applicable only when a ``group_by_field`` is specified. If set to true,
                                the percentage of each unique ``group_by_field`` value is calculated for each ``sum_nearby_layer`` feature.

                                The default is False.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    output_name                 Optional string or :class:`~arcgis.features.FeatureLayer`. Existing
                                feature layer will cause the new layer to be appended to the Feature Service.
                                If overwrite is True in context, new layer will overwrite existing layer.
                                If output_name not indicated then new :class:`~arcgis.features.FeatureCollection` created.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    context                     Optional dict. Additional settings such as processing extent and output spatial reference.
                                For summarize_nearby, there are three settings.

                                - ``extent`` - a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                                - ``outSR`` - the output features will be projected into the output spatial reference referred to by the `wkid`.
                                - ``overwrite`` - if True, then the feature layer in output_name will be overwritten with new feature layer. Available for ArcGIS Online or Enterprise 10.9.1+

                                    .. code-block:: python

                                        # Example Usage
                                        context = {"extent": {"xmin": 3164569.408035,
                                                            "ymin": -9187921.892449,
                                                            "xmax": 3174104.927313,
                                                            "ymax": -9175500.875353,
                                                            "spatialReference":{"wkid":102100,"latestWkid":3857}},
                                                    "outSR": {"wkid": 3857},
                                                    "overwrite": True}
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    estimate                    Optional boolean. Returns the number of credit for the operation.
    -------------------------   --------------------------------------------------------------------------------------------------------------------
    future                      Optional boolean. If True, a future object will be returned and the process
                                will not wait for the task to complete. The default is False, which means wait for results.
    =========================   ====================================================================================================================

    :return:
        result_layer : :class:`~arcgis.features.FeatureLayer` if output_name is specified, else :class:`~arcgis.features.FeatureCollection` dictionary.

        dict with the following keys:

        "result_layer" : layer (:class:`~arcgis.features.FeatureCollection`)

        "group_by_summary" : layer (:class:`~arcgis.features.FeatureCollection`)

    .. code-block:: python

         # USAGE EXAMPLE: To find hospital facilities that are within 5 miles of a school.
         summarize_nearby(sum_nearby_layer=item2.layers[0],
                          summary_layer=item1.layers[0],
                          near_type='StraightLine',
                          distances=[5],
                          units='Miles',
                          time_zone_for_time_of_day='GeoLocal',
                          return_boundaries=False,
                          sum_shape=True,
                          shape_units=None,
                          output_name='nearest hospitals to schools')
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    kwargs = {
        "sum_nearby_layer": sum_nearby_layer,
        "summary_layer": summary_layer,
        "near_type": near_type,
        "distances": distances,
        "units": units,
        "time_of_day": time_of_day,
        "time_zone_for_time_of_day": time_zone_for_time_of_day,
        "return_boundaries": return_boundaries,
        "sum_shape": sum_shape,
        "shape_units": shape_units,
        "summary_fields": summary_fields,
        "group_by_field": group_by_field,
        "percent_shape": percent_shape,
        "minority_majority": minority_majority,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "estimate": estimate,
        "future": future,
    }
    params = inspect_function_inputs(
        fn=gis._tools.featureanalysis.summarize_nearby, **kwargs
    )
    if isinstance(near_type, str):
        if near_type != "StraightLine":
            route_service = network.RouteLayer(
                gis.properties.helperServices.route.url, gis=gis
            )
            near_type = [
                i
                for i in route_service.retrieve_travel_modes()["supportedTravelModes"]
                if i["name"] == near_type
            ][0]
            params["near_type"] = near_type
    return gis._tools.featureanalysis.summarize_nearby(**params)


# --------------------------------------------------------------------------
def summarize_center_and_dispersion(
    analysis_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    summarize_type: list[str] = ["CentralFeature"],
    ellipse_size: Optional[str] = None,
    weight_field: Optional[str] = None,
    group_field: Optional[str] = None,
    output_name: Optional[Union[str, FeatureLayer]] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    estimate: bool = False,
    future: bool = False,
):
    """
    .. image:: _static/images/summarize_center_and_dispersion/summarize_center_and_dispersion.png

    The ``summarize_center_and_dispersion`` method finds central features and directional distributions. It can be used to answer questions such as:

    * Where is the center?
    * Which feature is the most accessible from all other features?
    * How dispersed, compact, or integrated are the features?
    * Are there directional trends?

    ====================    ============================================================================================
    **Parameter**            **Description**
    --------------------    --------------------------------------------------------------------------------------------
    analysis_layer          Required :class:`~arcgis.features.FeatureLayer` . The point, line, or polygon features to be analyzed. See :ref:`Feature Input<FeatureInput>`.
    --------------------    --------------------------------------------------------------------------------------------
    summarize_type          Required list of strings. The method with which to summarize the ``analysis_layer``.

                            Choice list: ["CentralFeature", "MeanCenter", "MedianCenter", "Ellipse"]
    --------------------    --------------------------------------------------------------------------------------------
    ellipse_size            Optional string. The size of the output ellipse in standard deviations.

                            Choice list: ['1 standard deviations', '2 standard deviations', '3 standard deviations']

                            The default ellipse size is '1 standard deviations'.
    --------------------    --------------------------------------------------------------------------------------------
    weight_field            Optional field. A numeric field in the ``analysis_layer`` to be used to
                            weight locations according to their relative importance.
    --------------------    --------------------------------------------------------------------------------------------
    group_field             Optional field. The field used to group features for separate directional
                            distribution calculations. The ``group_field`` can be of
                            integer, date, or string type.
    --------------------    --------------------------------------------------------------------------------------------
    output_name             Optional string or :class:`~arcgis.features.FeatureLayer`. Existing
                            feature layer will cause the new layer to be appended to the Feature Service.
                            If overwrite is True in context, new layer will overwrite existing layer.
                            If output_name not indicated then new :class:`~arcgis.features.FeatureCollection` created.
    --------------------    --------------------------------------------------------------------------------------------
    context                 Optional dict. Additional settings such as processing extent and output spatial reference.
                            For summarize_center_and_dispersion, there are three settings.

                            - ``extent`` - a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                            - ``outSR`` - the output features will be projected into the output spatial reference referred to by the `wkid`.
                            - ``overwrite`` - if True, then the feature layer in output_name will be overwritten with new feature layer. Available for ArcGIS Online or Enterprise 11+

                                .. code-block:: python

                                    # Example Usage
                                    context = {"extent": {"xmin": 3164569.408035,
                                                        "ymin": -9187921.892449,
                                                        "xmax": 3174104.927313,
                                                        "ymax": -9175500.875353,
                                                        "spatialReference":{"wkid":102100,"latestWkid":3857}},
                                                "outSR": {"wkid": 3857},
                                                "overwrite": True}
    --------------------    --------------------------------------------------------------------------------------------
    estimate                Optional boolean. If True, the number of credits to run the operation will be returned.
    --------------------    --------------------------------------------------------------------------------------------
    future                  Optional boolean. If True, a future object will be returned and the process
                            will not wait for the task to complete. The default is False, which means wait for results.
    ====================    ============================================================================================

    :return: list of items if ``output_name`` is supplied else, a Python dictionary with the following keys:

        | "central_feature_result_layer" : layer (:class:`~arcgis.features.FeatureCollection`)
        | "mean_feature_result_layer" : layer (:class:`~arcgis.features.FeatureCollection`)
        | "median_feature_result_layer" : layer (:class:`~arcgis.features.FeatureCollection`)
        | "ellipse_feature_result_layer" : layer (:class:`~arcgis.features.FeatureCollection`)

    .. code-block:: python

        # USAGE EXAMPLE: To find central features and mean center of earthquake over past months.
        central_features = summarize_center_and_dispersion(analysis_layer=earthquakes,
                                                           summarize_type=["CentralFeature","MeanCenter"],
                                                           ellipse_size='2 standard deviations',
                                                           weight_field='mag',
                                                           group_field='magType',
                                                           output_name='find central features and mean center of earthquake over past months')

    """

    gis = _arcgis.env.active_gis if gis is None else gis
    kwargs = {
        "analysis_layer": analysis_layer,
        "summarize_type": summarize_type,
        "ellipse_size": ellipse_size,
        "weight_field": weight_field,
        "group_field": group_field,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "estimate": estimate,
        "future": future,
    }
    params = inspect_function_inputs(
        fn=gis._tools.featureanalysis._tbx.summarize_center_and_dispersion,
        **kwargs,
    )
    return gis._tools.featureanalysis.summarize_center_and_dispersion(**params)


# --------------------------------------------------------------------------
def summarize_within(
    sum_within_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    summary_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    sum_shape: bool = True,
    shape_units: Optional[str] = None,
    summary_fields: Optional[list[str]] = [],
    group_by_field: Optional[str] = None,
    minority_majority: bool = False,
    percent_shape: bool = False,
    output_name: Optional[Union[str, FeatureLayer]] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    estimate: bool = False,
    future: bool = False,
    bin_type: str = "Square",
    bin_size: Optional[float] = None,
    bin_size_unit: Optional[str] = None,
):
    """
    .. image:: _static/images/summarize_within/summarize_within.png

    The ``summarize_within`` method finds the point, line, or polygon features (or portions of these features)
    that are within the boundaries of polygons in another layer. For example:

    * Given a layer of watershed boundaries and a layer of land-use boundaries by land-use type, calculate total acreage of land-use type for each watershed.
    * Given a layer of parcels in a county and a layer of city boundaries, summarize the average value of vacant parcels within each city boundary.
    * Given a layer of counties and a layer of roads, summarize the total mileage of roads by road type within each county.

    You can think of ``summarize_within`` as taking two layers and stacking them on top of each other.
    One of the layers, the ``sum_within_layer`` must be a polygon layer, and imagine that these polygon
    boundaries are all colored red. The other layer, the ``summary_layer``, can be any feature type point,
    line, or polygon. After stacking these layers on top of each other, you peer down through the stack
    and count the number of features in the ``summary_layer`` that fall within the polygons with the red
    boundaries (the ``sum_within_layer``). Not only can you count the number of features, you can calculate
    simple statistics about the attributes of the features in the ``summary_layer``, such as sum, mean, minimum, maximum, and so on.

    =====================================   =========================================================
    **Parameter**                            **Description**
    -------------------------------------   ---------------------------------------------------------
    sum_within_layer                        Required :class:`~arcgis.features.FeatureLayer` . The polygon features. Features, or
                                            portions of features, in the ``summary_layer`` (below) that fall within
                                            the boundaries of these polygons will be summarized. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------   ---------------------------------------------------------
    summary_layer                           Required :class:`~arcgis.features.FeatureLayer` . Point, line, or polygon features that will be summarized for each polygon in the ``sum_within_layer``.
                                            See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------   ---------------------------------------------------------
    sum_shape                               Optional boolean. A boolean value that instructs the task to calculate statistics
                                            based on shape type of the ``summary_layer``, such as the length of lines or areas of
                                            polygons of the ``summary_layer`` within each polygon in ``sum_within_layer``.

                                            The default is True.
    -------------------------------------   ---------------------------------------------------------
    shape_units                             Optional string. Specify units to summarize the length or areas when ``sum_shape`` is set to true. Units is not required to summarize
                                            points.

                                            * When ``summary_layer`` contains polygons: ['Acres', 'Hectares', 'SquareMeters', 'SquareKilometers', 'SquareMiles', 'SquareYards', 'SquareFeet']

                                            * When ``summary_layer`` contains lines: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']
    -------------------------------------   ---------------------------------------------------------
    summary_fields                          Optional list of strings. A list of field names and statistical summary type that you wish
                                            to calculate for all features in the ``summary_layer`` that are within each polygon in the ``sum_within_layer`` .

                                            Example: ["fieldname1 summary", "fieldname2 summary"]
    -------------------------------------   ---------------------------------------------------------
    group_by_field                          Optional string. This is a field of the ``summary_layer`` features that you can use to calculate statistics separately
                                            for each unique attribute value. For example, suppose the ``sum_within_layer`` contains city boundaries and
                                            the ``summary_layer`` features are parcels. One of the fields of the parcels is Status which contains
                                            two values: VACANT and OCCUPIED. To calculate the total area of vacant and occupied parcels within the
                                            boundaries of cities, use Status as the ``group_by_field`` field.
    -------------------------------------   ---------------------------------------------------------
    minority_majority                       Optional boolean. This boolean parameter is applicable only when a ``group_by_field`` is specified.
                                            If true, the minority (least dominant) or the majority (most dominant) attribute values for each group
                                            field are calculated. Two new fields are added to the ``result_layer`` prefixed with `Majority_` and `Minority_`.

                                            The default is False.
    -------------------------------------   ---------------------------------------------------------
    percent_shape                           Optional boolean. This Boolean parameter is applicable only when a ``group_by_field`` is specified.
                                            If set to true, the percentage of each unique ``group_by_field`` value is calculated for
                                            each ``sum_within_layer`` polygon.

                                            The default is False.
    -------------------------------------   ---------------------------------------------------------
    output_name                             Optional string or :class:`~arcgis.features.FeatureLayer`. Existing
                                            feature layer will cause the new layer to be appended to the Feature Service.
                                            If overwrite is True in context, new layer will overwrite existing layer.
                                            If output_name not indicated then new :class:`~arcgis.features.FeatureCollection` created.
    -------------------------------------   ---------------------------------------------------------
    context                                 Optional dict. Additional settings such as processing extent and output spatial reference.
                                            For summarize_within, there are three settings.

                                            - ``extent`` - a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                                            - ``outSR`` - the output features will be projected into the output spatial reference referred to by the `wkid`.
                                            - ``overwrite`` - if True, then the feature layer in output_name will be overwritten with new feature layer. Available for ArcGIS Online or Enterprise 10.9.1+

                                                .. code-block:: python

                                                    # Example Usage
                                                    context = {"extent": {"xmin": 3164569.408035,
                                                                        "ymin": -9187921.892449,
                                                                        "xmax": 3174104.927313,
                                                                        "ymax": -9175500.875353,
                                                                        "spatialReference":{"wkid":102100,"latestWkid":3857}},
                                                                "outSR": {"wkid": 3857},
                                                                "overwrite": True}
    -------------------------------------   ---------------------------------------------------------
    estimate                                Optional boolean. If True, the number of credits to run the operation will be returned.
    -------------------------------------   ---------------------------------------------------------
    future                                  Optional boolean. If True, a future object will be returned and the process
                                            will not wait for the task to complete. The default is False, which means wait for results.
    -------------------------------------   ---------------------------------------------------------
    bin_type                                Required string. The type of bin used to calculate density.

                                            Choice list: ['Hexagon', 'Square'].
    -------------------------------------   ---------------------------------------------------------
    bin_size                                Required float. The distance for the bins that the ``input_layer`` will be analyzed using.
                                            When generating bins, for Square, the number and units specified determine the
                                            height and length of the square. For ``Hexagon``, the number and units specified
                                            determine the distance between parallel sides.
    -------------------------------------   ---------------------------------------------------------
    bin_size_unit                           Required string. The distance unit for the bins for which the density will be calculated.
                                            The linear unit to be used with the value specified in ``bin_size``.

                                            The default is 'Meters'.
    =====================================   =========================================================

    :return:
        result_layer : :class:`~arcgis.features.FeatureLayer` if output_name is specified, else :class:`~arcgis.features.FeatureCollection` dictionary.

        dict with the following keys:

        "result_layer" : layer (FeatureCollection)

        "group_by_summary" : layer (FeatureCollection)

    .. code-block:: python

        # USAGE EXAMPLE: To summarize traffic accidents within each county and group them by the day of accident.
        acc_within_county = summarize_within(sum_within_layer=boundaries,
                                             summary_layer=collision_lyr,
                                             sum_shape=True,
                                             group_by_field='Day',
                                             minority_majority=True,
                                             percent_shape=True,
                                             output_name='summarize accidents within each county',
                                             context={"extent":{"xmin":-13160690.837046918,"ymin":4041586.5461609075,"xmax":-13132466.464352652,"ymax":4058001.397985127,"spatialReference":{"wkid":102100,"latestWkid":3857}}})
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    kwargs = {
        "sum_within_layer": sum_within_layer,
        "summary_layer": summary_layer,
        "sum_shape": sum_shape,
        "shape_units": shape_units,
        "summary_fields": summary_fields,
        "group_by_field": group_by_field,
        "minority_majority": minority_majority,
        "percent_shape": percent_shape,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "estimate": estimate,
        "future": future,
        "bin_type": bin_type,
        "bin_size": bin_size,
        "bin_size_unit": bin_size_unit,
    }
    params = inspect_function_inputs(
        fn=gis._tools.featureanalysis._tbx.summarize_within, **kwargs
    )
    if not estimate is None:
        params["estimate"] = estimate
    return gis._tools.featureanalysis.summarize_within(**params)


# --------------------------------------------------------------------------
def join_features(
    target_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    join_layer: Union[
        Item,
        FeatureCollection,
        FeatureLayer,
        FeatureLayerCollection,
        str,
        dict[str, Any],
    ],
    spatial_relationship: Optional[str] = None,
    spatial_relationship_distance: Optional[float] = None,
    spatial_relationship_distance_units: Optional[str] = None,
    attribute_relationship: Optional[list[dict[str, Any]]] = None,
    join_operation: str = """JoinOneToOne""",
    summary_fields: Optional[list[dict[str[Any]]]] = None,
    output_name: Optional[Union[str, FeatureLayer]] = None,
    context: Optional[dict[str, Any]] = None,
    gis: Optional[GIS] = None,
    estimate: bool = False,
    future: bool = False,
    join_type: str = "INNER",
    records_to_match: Optional[dict[str, Any]] = None,
):
    """
    .. image:: _static/images/join_features/join_features.png

    The ``join_features`` method works with two layers and joins the attributes
    from one feature to another based on spatial and attribute relationships.

    ============================================================================================    =================================================================================================================================
    **Parameter**                                                                                    **Description**
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    target_layer                                                                                    Required layer. The point, line, polygon or table layer that will have attributes from
                                                                                                    the ``join_layer`` appended to its table. See :ref:`Feature Input<FeatureInput>`.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    join_layer                                                                                      Required layer. The point, line, polygon or table layer that will be joined to the ``target_layer``. See :ref:`Feature Input<FeatureInput>`.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship                                                                            Required string if not table layers. Defines the spatial relationship used to spatially join features.

                                                                                                    Choice list: ['identicalto', 'intersects', 'completelycontains', 'completelywithin', 'withindistance']
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship_distance                                                                   Optional float. A float value used for the search distance to determine if the target features are near or within a
    (Required if ``spatial_relationship`` is withindistance)                                        specified distance of the join features.
                                                                                                    This is only applied if Within a distance of is the selected ``spatial_relationship``.
                                                                                                    You can only enter a single distance value. The units of the distance values are supplied by the
                                                                                                    ``spatial_relationship_distance_units`` parameter.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    spatial_relationship_distance_units                                                             Optional string. The linear unit to be used with the distance value specified in ``spatial_relationship_distance``.
    (Required if ``spatial_relationship`` is withindistance)                                        Choice list: ['Miles', 'Yards', 'Feet', 'NauticalMiles', 'Meters', 'Kilometers']

                                                                                                    The default is 'Miles'.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    attribute_relationship                                                                          Optional list of dicts. Defines an attribute relationship used to join features. Features are matched when the field
                                                                                                    values in the join layer are equal to field values in the target layer.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    join_operation                                                                                  Optional string. A string representing the type of join that will be applied

                                                                                                    Choice list: ['JoinOneToOne', 'JoinOneToMany']

                                                                                                    * ``JoinOneToOne`` - If multiple join features are found that have the same relationships with asingle target feature, the attributes from the multiple join features will be aggregated using the specified summary statistics.
                                                                                                      For example, if a point target feature is found within two separate polygon join features, the attributes from the two polygons will be aggregated before being transferred to the output point feature class.
                                                                                                      If one polygon has an attribute value of 3 and the other has a value of 7, and a SummaryField of sum is selected, the aggregated value
                                                                                                      in the output feature class will be 10. There will always be a Count field calculated, with a value of 2, for the number of features specified. This is the default.

                                                                                                    * ``JoinOneToMany`` - If multiple join features are found that have the same relationship with a single target feature, the output feature class will contain multiple copies (records) of the target feature.
                                                                                                      For example, if a single point target feature is found within two separate polygon join features, the output feature class will contain two copies of the target feature:
                                                                                                      one record with the attributes of the first polygon, and another record with the attributes of the second polygon. There are no summary statistics calculated with this method.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    summary_fields                                                                                  Optional list of dicts. A list of field names and statistical summary types that you want to calculate. Note that the count is always returned by default.

                                                                                                    fieldName is the name of one of the numeric fields found in the input join layer.

                                                                                                    statisticType is one of the following:

                                                                                                    * ``SUM`` - Adds the total value of all the points in each polygon
                                                                                                    * ``MEAN`` - Calculates the average of all the points in each polygon
                                                                                                    * ``MIN`` - Finds the smallest value of all the points in each polygon
                                                                                                    * ``MAX`` - Finds the largest value of all the points in each polygon
                                                                                                    * ``STDDEV`` - Finds the standard deviation of all the points in each polygon
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    output_name                                                                                     Optional string or :class:`~arcgis.features.FeatureLayer`. Existing
                                                                                                    feature layer will cause the new layer to be appended to the Feature Service.
                                                                                                    If overwrite is True in context, new layer will overwrite existing layer.
                                                                                                    If output_name not indicated then new :class:`~arcgis.features.FeatureCollection` created.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    context                                                                                         Optional dict. Additional settings such as processing extent and output spatial reference.
                                                                                                    For join_features, there are three settings.

                                                                                                    - ``extent`` - a bounding box that defines the analysis area. Only those features in the input_layer that intersect the bounding box will be analyzed.
                                                                                                    - ``outSR`` - the output features will be projected into the output spatial reference referred to by the `wkid`.
                                                                                                    - ``overwrite`` - if True, then the feature layer in output_name will be overwritten with new feature layer. Available for ArcGIS Online or Enterprise 10.9.1+

                                                                                                        .. code-block:: python

                                                                                                            # Example Usage
                                                                                                            context = {"extent": {"xmin": 3164569.408035,
                                                                                                                                "ymin": -9187921.892449,
                                                                                                                                "xmax": 3174104.927313,
                                                                                                                                "ymax": -9175500.875353,
                                                                                                                                "spatialReference":{"wkid":102100,"latestWkid":3857}},
                                                                                                                        "outSR": {"wkid": 3857},
                                                                                                                        "overwrite": True}
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    estimate                                                                                        Optional boolean. If True, the number of credits to run the operation will be returned.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    future                                                                                          Optional boolean. If True, a future object will be returned and the process
                                                                                                    will not wait for the task to complete. The default is False, which means wait for results.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    join_type                                                                                       Optional String.  Determines the type of join performed on the datasets.  The allowed values are INNER or LEFT.
    --------------------------------------------------------------------------------------------    ---------------------------------------------------------------------------------------------------------------------------------
    records_to_match                                                                                Optional Dict. Defines how two features are joined.
                                                                                                    Example:
                                                                                                        | {"groupByFields":"",
                                                                                                        | "orderByFields":"objectid ASC",
                                                                                                        | "topCount":1}
    ============================================================================================    =================================================================================================================================

    :return: result_layer : :class:`~arcgis.features.FeatureLayer` if output_name is specified, else :class:`~arcgis.features.FeatureCollection`.

    .. code-block:: python

        USAGE EXAMPLE: To summarize traffic accidents within each parcel using spatial relationship.
        accident_count_in_each_parcel = join_features(target_layer=parcel_lyr,
                                                      join_layer=traffic_accidents_lyr,
                                                      spatial_relationship='intersects',
                                                      summary_fields=[{"statisticType": "Mean", "onStatisticField": "Population"},
                                                      output_name='join features',
                                                      context={"extent":{"xmin":-9375809.87305117,"ymin":4031882.3806860778,"xmax":-9370182.196843527,"ymax":4034872.9794178144,"spatialReference":{"wkid":102100,"latestWkid":3857}}}, )
    """
    kwargs = {
        "target_layer": target_layer,
        "join_layer": join_layer,
        "spatial_relationship": spatial_relationship,
        "spatial_relationship_distance": spatial_relationship_distance,
        "spatial_relationship_distance_units": spatial_relationship_distance_units,
        "attribute_relationship": attribute_relationship,
        "join_operation": join_operation,
        "summary_fields": summary_fields,
        "output_name": output_name,
        "context": context,
        "gis": gis,
        "future": future,
        "join_type": join_type,
        "records_to_match": records_to_match,
    }
    gis = _arcgis.env.active_gis if gis is None else gis
    params = inspect_function_inputs(
        fn=gis._tools.featureanalysis._tbx.join_features, **kwargs
    )
    if not estimate is None:
        params["estimate"] = estimate
    if "attribute_relationship" not in params:
        params["attribute_relationship"] = attribute_relationship
    if "summary_fields" not in params:
        params["summary_fields"] = summary_fields
    return gis._tools.featureanalysis.join_features(**params)
